#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Manage upstream ansible-core releases."""

from __future__ import annotations

import argparse
import contextlib
import dataclasses
import datetime
import enum
import functools
import gzip
import hashlib
import http.client
import inspect
import json
import math
import os
import pathlib
import re
import secrets
import shlex
import shutil
import stat
import subprocess
import sys
import tarfile
import tempfile
import typing as t
import urllib.error
import urllib.parse
import urllib.request
import venv
import webbrowser
import zipfile

import jinja2

from packaging.version import Version, InvalidVersion

# region CLI Framework


C = t.TypeVar("C", bound=t.Callable[..., None])


def path_to_str(value: t.Any) -> str:
    """Return the given value converted to a string suitable for use as a command line argument."""
    return f"{value}/" if isinstance(value, pathlib.Path) and value.is_dir() else str(value)


@t.overload
def run(*args: t.Any, env: dict[str, t.Any] | None, cwd: pathlib.Path | str, capture_output: t.Literal[True]) -> CompletedProcess: ...


@t.overload
def run(*args: t.Any, env: dict[str, t.Any] | None, cwd: pathlib.Path | str, capture_output: t.Literal[False]) -> None: ...


@t.overload
def run(*args: t.Any, env: dict[str, t.Any] | None, cwd: pathlib.Path | str, capture_output: bool) -> CompletedProcess | None: ...


@t.overload
def run(*args: t.Any, env: dict[str, t.Any] | None, cwd: pathlib.Path | str) -> None: ...


def run(
    *args: t.Any,
    env: dict[str, t.Any] | None,
    cwd: pathlib.Path | str,
    capture_output: bool = False,
) -> CompletedProcess | None:
    """Run the specified command."""
    args = [arg.relative_to(cwd) if isinstance(arg, pathlib.Path) else arg for arg in args]

    str_args = tuple(path_to_str(arg) for arg in args)
    str_env = {key: path_to_str(value) for key, value in env.items()} if env is not None else None

    display.show(f"--> {shlex.join(str_args)}", color=Display.CYAN)

    try:
        p = subprocess.run(str_args, check=True, text=True, env=str_env, cwd=cwd, capture_output=capture_output)
    except subprocess.CalledProcessError as ex:
        # improve type hinting and include stdout/stderr (if any) in the message
        raise CalledProcessError(
            message=str(ex),
            cmd=str_args,
            status=ex.returncode,
            stdout=ex.stdout,
            stderr=ex.stderr,
        ) from None

    if not capture_output:
        return None

    # improve type hinting
    return CompletedProcess(
        stdout=p.stdout,
        stderr=p.stderr,
    )


@contextlib.contextmanager
def suppress_when(error_as_warning: bool) -> t.Generator[None, None, None]:
    """Conditionally convert an ApplicationError in the provided context to a warning."""
    if error_as_warning:
        try:
            yield
        except ApplicationError as ex:
            display.warning(ex)
    else:
        yield


class ApplicationError(Exception):
    """A fatal application error which will be shown without a traceback."""


class CalledProcessError(Exception):
    """Results from a failed process."""

    def __init__(self, message: str, cmd: tuple[str, ...], status: int, stdout: str | None, stderr: str | None) -> None:
        if stdout and (stdout := stdout.strip()):
            message += f"\n>>> Standard Output\n{stdout}"

        if stderr and (stderr := stderr.strip()):
            message += f"\n>>> Standard Error\n{stderr}"

        super().__init__(message)

        self.cmd = cmd
        self.status = status
        self.stdout = stdout
        self.stderr = stderr


@dataclasses.dataclass(frozen=True)
class CompletedProcess:
    """Results from a completed process."""

    stdout: str
    stderr: str


class Display:
    """Display interface for sending output to the console."""

    CLEAR = "\033[0m"
    RED = "\033[31m"
    BLUE = "\033[34m"
    PURPLE = "\033[35m"
    CYAN = "\033[36m"

    def fatal(self, message: t.Any) -> None:
        """Print a fatal message to the console."""
        self.show(f"FATAL: {message}", color=self.RED)

    def warning(self, message: t.Any) -> None:
        """Print a warning message to the console."""
        self.show(f"WARNING: {message}", color=self.PURPLE)

    def show(self, message: t.Any, color: str | None = None) -> None:
        """Print a message to the console."""
        print(f"{color or self.CLEAR}{message}{self.CLEAR}", flush=True)


class CommandFramework:
    """
    Simple command line framework inspired by nox.

    Argument parsing is handled by argparse. Each function annotated with an instance of this class becomes a subcommand.
    Options are shared across all commands, and are defined by providing kwargs when creating an instance of this class.
    Options are only defined for commands which have a matching parameter.

    The name of each kwarg is the option name, which will be prefixed with `--` and with underscores converted to dashes.
    The value of each kwarg is passed as kwargs to ArgumentParser.add_argument. Passing None results in an internal only parameter.

    The following custom kwargs are recognized and are not passed to add_argument:

    name - Override the positional argument (option) passed to add_argument.
    exclusive - Put the argument in an exclusive group of the given name.
    """

    def __init__(self, **kwargs: dict[str, t.Any] | None) -> None:
        self.commands: list[t.Callable[..., None]] = []
        self.arguments = kwargs
        self.parsed_arguments: argparse.Namespace | None = None

    def __call__(self, func: C) -> C:
        """Register the decorated function as a CLI command."""
        self.commands.append(func)
        return func

    def run(self, *args: t.Callable[..., None], **kwargs) -> None:
        """Run the specified command(s), using any provided internal args."""
        for arg in args:
            self._run(arg, **kwargs)

    def main(self) -> None:
        """Main program entry point."""
        parser = argparse.ArgumentParser(description=__doc__)
        subparsers = parser.add_subparsers(metavar="COMMAND", required=True)

        for func in self.commands:
            func_parser = subparsers.add_parser(self._format_command_name(func), description=func.__doc__, help=func.__doc__)
            func_parser.set_defaults(func=func)

            exclusive_groups = {}
            signature = inspect.signature(func)

            for name in signature.parameters:
                if name not in self.arguments:
                    raise RuntimeError(f"The '{name}' argument, used by '{func.__name__}', has not been defined.")

                if (arguments := self.arguments.get(name)) is None:
                    continue  # internal use

                arguments = arguments.copy()
                exclusive = arguments.pop("exclusive", None)

                # noinspection PyProtectedMember, PyUnresolvedReferences
                command_parser: argparse._ActionsContainer

                if exclusive:
                    if exclusive not in exclusive_groups:
                        exclusive_groups[exclusive] = func_parser.add_mutually_exclusive_group()

                    command_parser = exclusive_groups[exclusive]
                else:
                    command_parser = func_parser

                if option_name := arguments.pop("name", None):
                    arguments.update(dest=name)
                else:
                    option_name = f"--{name.replace('_', '-')}"

                command_parser.add_argument(option_name, **arguments)

        try:
            # noinspection PyUnresolvedReferences
            import argcomplete
        except ImportError:
            pass
        else:
            argcomplete.autocomplete(parser)

        self.parsed_arguments = parser.parse_args()

        try:
            self.run(self.parsed_arguments.func)
        except ApplicationError as ex:
            display.fatal(ex)
            sys.exit(1)

    def _run(self, func: t.Callable[..., None], **kwargs) -> None:
        """Run the specified command, using any provided internal args."""
        signature = inspect.signature(func)
        func_args = {name: getattr(self.parsed_arguments, name) for name in signature.parameters if hasattr(self.parsed_arguments, name)}
        func_args.update({name: value for name, value in kwargs.items() if name in signature.parameters})
        printable_args = ", ".join(f"{name}={repr(value)}" for name, value in func_args.items())
        label = f"{self._format_command_name(func)}({printable_args})"

        display.show(f"==> {label}", color=Display.BLUE)

        try:
            func(**func_args)
        except BaseException:
            display.show(f"!!! {label}", color=Display.RED)
            raise

        display.show(f"<== {label}", color=Display.BLUE)

    @staticmethod
    def _format_command_name(func: t.Callable[..., None]) -> str:
        """Return the friendly name of the given command."""
        return func.__name__.replace("_", "-")


display = Display()


# endregion
# region Data Classes


@dataclasses.dataclass(frozen=True)
class GitHubRelease:
    """Details required to create a GitHub release."""

    user: str
    repo: str
    tag: str
    target: str
    title: str
    body: str
    pre_release: bool


@dataclasses.dataclass(frozen=True)
class PullRequest:
    """Details required to create a pull request."""

    upstream_user: str
    upstream_repo: str
    upstream_branch: str
    user: str
    repo: str
    branch: str
    title: str
    body: str


@dataclasses.dataclass(frozen=True)
class Remote:
    """Details about a git remote."""

    name: str
    user: str
    repo: str


@dataclasses.dataclass(frozen=True)
class Remotes:
    """Details about git removes."""

    fork: Remote
    upstream: Remote


@dataclasses.dataclass(frozen=True)
class GitState:
    """Details about the state of the git repository."""

    remotes: Remotes
    branch: str | None
    commit: str


@dataclasses.dataclass(frozen=True)
class ReleaseArtifact:
    """Information about a release artifact on PyPI."""

    package_type: str
    package_label: str
    url: str
    size: int
    digest: str
    digest_algorithm: str


# endregion
# region Utilities


SCRIPT_DIR = pathlib.Path(__file__).parent.resolve()
CHECKOUT_DIR = SCRIPT_DIR.parent

ANSIBLE_LIB_DIR = CHECKOUT_DIR / "lib"
ANSIBLE_DIR = ANSIBLE_LIB_DIR / "ansible"
ANSIBLE_BIN_DIR = CHECKOUT_DIR / "bin"
ANSIBLE_RELEASE_FILE = ANSIBLE_DIR / "release.py"
ANSIBLE_REQUIREMENTS_FILE = CHECKOUT_DIR / "requirements.txt"
ANSIBLE_CHANGELOG_REQUIREMENTS_FILE = CHECKOUT_DIR / "test/lib/ansible_test/_data/requirements/sanity.changelog.txt"
ANSIBLE_PYPROJECT_TOML_FILE = CHECKOUT_DIR / "pyproject.toml"

DIST_DIR = CHECKOUT_DIR / "dist"
VENV_DIR = DIST_DIR / ".venv" / "release"

CHANGELOGS_DIR = CHECKOUT_DIR / "changelogs"
CHANGELOGS_FRAGMENTS_DIR = CHANGELOGS_DIR / "fragments"

ANSIBLE_VERSION_PATTERN = re.compile("^__version__ = '(?P<version>.*)'$", re.MULTILINE)
ANSIBLE_VERSION_FORMAT = "__version__ = '{version}'"

DIGEST_ALGORITHM = "sha256"

# These endpoint names match those defined as defaults in twine.
# See: https://github.com/pypa/twine/blob/9c2c0a1c535155931c3d879359330cb836950c6a/twine/utils.py#L82-L85
PYPI_ENDPOINTS = dict(
    pypi="https://pypi.org/pypi",
    testpypi="https://test.pypi.org/pypi",
)

PIP_ENV = dict(
    PIP_REQUIRE_VIRTUALENV="yes",
    PIP_DISABLE_PIP_VERSION_CHECK="yes",
)


class VersionMode(enum.Enum):
    """How to handle the ansible-core version."""

    DEFAULT = enum.auto()
    """Do not allow development versions. Do not allow post release versions."""
    STRIP_POST = enum.auto()
    """Do not allow development versions. Strip the post release from the version if present."""
    REQUIRE_POST = enum.auto()
    """Do not allow development versions. Require a post release version."""
    REQUIRE_DEV_POST = enum.auto()
    """Require a development or post release version."""
    ALLOW_DEV_POST = enum.auto()
    """Allow development and post release versions."""

    def apply(self, version: Version) -> Version:
        """Apply the mode to the given version and return the result."""
        original_version = version

        release_component_count = 3

        if len(version.release) != release_component_count:
            raise ApplicationError(f"Version {version} contains {version.release} release components instead of {release_component_count}.")

        if version.epoch:
            raise ApplicationError(f"Version {version} contains an epoch component: {version.epoch}")

        if version.local is not None:
            raise ApplicationError(f"Version {version} contains a local component: {version.local}")

        if version.is_devrelease and version.is_postrelease:
            raise ApplicationError(f"Version {version} is a development and post release version.")

        if self == VersionMode.ALLOW_DEV_POST:
            return version

        if self == VersionMode.REQUIRE_DEV_POST:
            if not version.is_devrelease and not version.is_postrelease:
                raise ApplicationError(f"Version {version} is not a development or post release version.")

            return version

        if version.is_devrelease:
            raise ApplicationError(f"Version {version} is a development release: {version.dev}")

        if self == VersionMode.STRIP_POST:
            if version.is_postrelease:
                version = Version(str(version).removesuffix(f".post{version.post}"))
                display.warning(f"Using version {version} by stripping the post release suffix from version {original_version}.")

            return version

        if self == VersionMode.REQUIRE_POST:
            if not version.is_postrelease:
                raise ApplicationError(f"Version {version} is not a post release version.")

            return version

        if version.is_postrelease:
            raise ApplicationError(f"Version {version} is a post release.")

        if self == VersionMode.DEFAULT:
            return version

        raise NotImplementedError(self)


@t.overload
def git(*args: t.Any, capture_output: t.Literal[True]) -> CompletedProcess: ...


@t.overload
def git(*args: t.Any, capture_output: t.Literal[False]) -> None: ...


@t.overload
def git(*args: t.Any) -> None: ...


def git(*args: t.Any, capture_output: bool = False) -> CompletedProcess | None:
    """Run the specified git command."""
    return run("git", *args, env=None, cwd=CHECKOUT_DIR, capture_output=capture_output)


def get_commit(rev: str | None = None) -> str:
    """Return the commit associated with the given rev, or HEAD if no rev is given."""
    try:
        return git("rev-parse", "--quiet", "--verify", "--end-of-options", f"{rev or 'HEAD'}^{{commit}}", capture_output=True).stdout.strip()
    except CalledProcessError as ex:
        if ex.status == 1 and not ex.stdout and not ex.stderr:
            raise ApplicationError(f"Could not find commit: {rev}") from None

        raise


def prepare_pull_request(version: Version, branch: str, title: str, add: t.Iterable[pathlib.Path | str], allow_stale: bool) -> PullRequest:
    """Return pull request parameters using the provided details."""
    git_state = get_git_state(version, allow_stale)

    if not git("status", "--porcelain", "--untracked-files=no", capture_output=True).stdout.strip():
        raise ApplicationError("There are no changes to commit. Did you skip a step?")

    upstream_branch = get_upstream_branch(version)
    body = create_pull_request_body(title)

    git("checkout", "-b", branch)
    git("add", *add)
    git("commit", "-m", title)
    git("push", "--set-upstream", git_state.remotes.fork.name, branch)
    git("checkout", git_state.branch or git_state.commit)
    git("branch", "-d", branch)

    pr = PullRequest(
        upstream_user=git_state.remotes.upstream.user,
        upstream_repo=git_state.remotes.upstream.repo,
        upstream_branch=upstream_branch,
        user=git_state.remotes.fork.user,
        repo=git_state.remotes.fork.repo,
        branch=branch,
        title=title,
        body=body,
    )

    return pr


def create_github_release(release: GitHubRelease) -> None:
    """Open a browser tab for creating the given GitHub release."""
    # See: https://docs.github.com/en/repositories/releasing-projects-on-github/automation-for-release-forms-with-query-parameters

    params = dict(
        tag=release.tag,
        target=release.target,
        title=release.title,
        body=release.body,
        prerelease=1 if release.pre_release else 0,
    )

    query_string = urllib.parse.urlencode(params)
    url = f"https://github.com/{release.user}/{release.repo}/releases/new?{query_string}"

    display.show("Opening release creation page in new tab using default browser ...")
    webbrowser.open_new_tab(url)


def create_pull_request(pr: PullRequest) -> None:
    """Open a browser tab for creating the given pull request."""
    # See: https://docs.github.com/en/pull-requests/collaborating-with-pull-requests/proposing-changes-to-your-work-with-pull-requests/using-query-parameters-to-create-a-pull-request  # noqa

    params = dict(
        quick_pull=1,
        title=pr.title,
        body=pr.body,
    )

    query_string = urllib.parse.urlencode(params)
    url = f"https://github.com/{pr.upstream_user}/{pr.upstream_repo}/compare/{pr.upstream_branch}...{pr.user}:{pr.repo}:{pr.branch}?{query_string}"

    display.show("Opening pull request in new tab using default browser ...")
    webbrowser.open_new_tab(url)


def create_pull_request_body(title: str) -> str:
    """Return a simple pull request body created from the given title."""
    body = f"""
##### SUMMARY

{title}

##### ISSUE TYPE

Feature Pull Request
"""

    return body.lstrip()


def get_remote(name: str, push: bool) -> Remote:
    """Return details about the specified remote."""
    remote_url = git("remote", "get-url", *(["--push"] if push else []), name, capture_output=True).stdout.strip()
    remote_match = re.search(r"[@/]github[.]com[:/](?P<user>[^/]+)/(?P<repo>[^.]+)(?:[.]git)?$", remote_url)

    if not remote_match:
        raise RuntimeError(f"Unable to identify the user and repo in the '{name}' remote: {remote_url}")

    remote = Remote(
        name=name,
        user=remote_match.group("user"),
        repo=remote_match.group("repo"),
    )

    return remote


@functools.cache
def get_remotes() -> Remotes:
    """Return details about the remotes we need to use."""
    # assume the devel branch has its upstream remote pointing to the user's fork
    fork_remote_name = git("branch", "--list", "devel", "--format=%(upstream:remotename)", capture_output=True).stdout.strip()

    if not fork_remote_name:
        raise ApplicationError("Could not determine the remote for your fork of Ansible.")

    display.show(f"Detected '{fork_remote_name}' as the remote for your fork of Ansible.")

    # assume there is only one ansible org remote, which would allow release testing using another repo in the same org without special configuration
    all_remotes = git("remote", "-v", capture_output=True).stdout.strip().splitlines()
    ansible_remote_names = set(line.split()[0] for line in all_remotes if re.search(r"[@/]github[.]com[:/]ansible/", line))

    if not ansible_remote_names:
        raise ApplicationError(f"Could not determine the remote which '{fork_remote_name}' was forked from.")

    if len(ansible_remote_names) > 1:
        raise ApplicationError(f"Found multiple candidates for the remote from which '{fork_remote_name}' was forked from: {', '.join(ansible_remote_names)}")

    upstream_remote_name = ansible_remote_names.pop()

    display.show(f"Detected '{upstream_remote_name}' as the remote from which '{fork_remote_name}' was forked from.")

    if fork_remote_name == upstream_remote_name:
        raise ApplicationError("The remote for your fork of Ansible cannot be the same as the remote from which it was forked.")

    remotes = Remotes(
        fork=get_remote(fork_remote_name, push=True),
        upstream=get_remote(upstream_remote_name, push=False),
    )

    return remotes


def get_upstream_branch(version: Version) -> str:
    """Return the upstream branch name for the given version."""
    return f"stable-{version.major}.{version.minor}"


def get_git_state(version: Version, allow_stale: bool) -> GitState:
    """Return information about the current state of the git repository."""
    remotes = get_remotes()
    upstream_branch = get_upstream_branch(version)

    git("fetch", remotes.upstream.name, upstream_branch)

    upstream_ref = f"{remotes.upstream.name}/{upstream_branch}"
    upstream_commit = get_commit(upstream_ref)

    commit = get_commit()

    if commit != upstream_commit:
        with suppress_when(allow_stale):
            raise ApplicationError(f"The current commit ({commit}) does not match {upstream_ref} ({upstream_commit}).")

    branch = git("branch", "--show-current", capture_output=True).stdout.strip() or None

    state = GitState(
        remotes=remotes,
        branch=branch,
        commit=commit,
    )

    return state


@functools.cache
def ensure_venv(requirements_content: str) -> dict[str, t.Any]:
    """Ensure the release venv is ready and return the env vars needed to use it."""
    requirements_hash = hashlib.sha256(requirements_content.encode()).hexdigest()[:8]

    python_version = ".".join(map(str, sys.version_info[:2]))

    venv_dir = VENV_DIR / python_version / requirements_hash
    venv_bin_dir = venv_dir / "bin"
    venv_requirements_file = venv_dir / "requirements.txt"
    venv_marker_file = venv_dir / "marker.txt"

    env = os.environ.copy()
    env.pop("PYTHONPATH", None)  # avoid interference from ansible being injected into the environment
    env.update(
        PATH=os.pathsep.join((str(venv_bin_dir), env["PATH"])),
    )

    if not venv_marker_file.exists():
        display.show(f"Creating a Python {python_version} virtual environment ({requirements_hash}) ...")

        if venv_dir.exists():
            shutil.rmtree(venv_dir)

        venv.create(venv_dir, with_pip=True)

        venv_requirements_file.write_text(requirements_content)

        run("pip", "install", "-r", venv_requirements_file, env=env | PIP_ENV, cwd=CHECKOUT_DIR)

        venv_marker_file.touch()

    return env


def get_pypi_project(repository: str, project: str, version: Version | None = None) -> dict[str, t.Any]:
    """Return the project JSON from PyPI for the specified repository, project and version (optional)."""
    endpoint = PYPI_ENDPOINTS[repository]

    if version:
        url = f"{endpoint}/{project}/{version}/json"
    else:
        url = f"{endpoint}/{project}/json"

    opener = urllib.request.build_opener()
    response: http.client.HTTPResponse

    try:
        with opener.open(url) as response:
            data = json.load(response)
    except urllib.error.HTTPError as ex:
        if version:
            target = f'{project!r} version {version}'
        else:
            target = f'{project!r}'

        if ex.status == http.HTTPStatus.NOT_FOUND:
            raise ApplicationError(f"Could not find {target} on PyPI.") from None

        raise RuntimeError(f"Failed to get {target} from PyPI.") from ex

    return data


def get_ansible_version(version: str | None = None, /, commit: str | None = None, mode: VersionMode = VersionMode.DEFAULT) -> Version:
    """Parse and return the current ansible-core version, the provided version or the version from the provided commit."""
    if version and commit:
        raise ValueError("Specify only one of: version, commit")

    if version:
        source = ""
    else:
        if commit:
            current = git("show", f"{commit}:{ANSIBLE_RELEASE_FILE.relative_to(CHECKOUT_DIR)}", capture_output=True).stdout
        else:
            current = ANSIBLE_RELEASE_FILE.read_text()

        if not (match := ANSIBLE_VERSION_PATTERN.search(current)):
            raise RuntimeError("Failed to get the ansible-core version.")

        version = match.group("version")
        source = f" in '{ANSIBLE_RELEASE_FILE}'"

    try:
        parsed_version = Version(version)
    except InvalidVersion:
        raise ApplicationError(f"Invalid version{source}: {version}") from None

    parsed_version = mode.apply(parsed_version)

    return parsed_version


def get_next_version(version: Version, /, final: bool = False, pre: str | None = None, mode: VersionMode = VersionMode.DEFAULT) -> Version:
    """Return the next version after the specified version."""

    # TODO: consider using development versions instead of post versions after a release is published

    pre = pre or ""
    micro = version.micro

    if version.is_devrelease:
        # The next version of a development release is the same version without the development component.
        if final:
            pre = ""
        elif not pre and version.pre is not None:
            pre = f"{version.pre[0]}{version.pre[1]}"
        elif not pre:
            pre = "b1"  # when there is no existing pre and none specified, advance to b1

    elif version.is_postrelease:
        # The next version of a post release is the next pre-release *or* micro release component.
        if final:
            pre = ""
        elif not pre and version.pre is not None:
            pre = f"{version.pre[0]}{version.pre[1] + 1}"
        elif not pre:
            pre = "rc1"  # when there is no existing pre and none specified, advance to rc1

        if version.pre is None:
            micro = version.micro + 1
    else:
        raise ApplicationError(f"Version {version} is not a development or post release version.")

    version = f"{version.major}.{version.minor}.{micro}{pre}"

    return get_ansible_version(version, mode=mode)


def check_ansible_version(current_version: Version, requested_version: Version) -> None:
    """Verify the requested version is valid for the current version."""
    if requested_version.release[:2] != current_version.release[:2]:
        raise ApplicationError(f"Version {requested_version} does not match the major and minor portion of the current version: {current_version}")

    if requested_version < current_version:
        raise ApplicationError(f"Version {requested_version} is older than the current version: {current_version}")

    # TODO: consider additional checks to avoid mistakes when incrementing the release version


def set_ansible_version(current_version: Version, requested_version: Version) -> None:
    """Set the current ansible-core version."""
    check_ansible_version(current_version, requested_version)

    if requested_version == current_version:
        return

    display.show(f"Updating version {current_version} to {requested_version} ...")

    current = ANSIBLE_RELEASE_FILE.read_text()
    updated = ANSIBLE_VERSION_PATTERN.sub(ANSIBLE_VERSION_FORMAT.format(version=requested_version), current)

    if current == updated:
        raise RuntimeError("Failed to set the ansible-core version.")

    ANSIBLE_RELEASE_FILE.write_text(updated)


def get_latest_setuptools_version() -> Version:
    """Return the latest setuptools version found on PyPI."""
    data = get_pypi_project('pypi', 'setuptools')
    version = Version(data['info']['version'])

    return version


def set_setuptools_upper_bound(requested_version: Version) -> None:
    """Set the upper bound on setuptools in pyproject.toml."""
    current = ANSIBLE_PYPROJECT_TOML_FILE.read_text()
    pattern = re.compile(r'^(?P<begin>requires = \["setuptools >= )(?P<lower>[^,]+)(?P<middle>, <= )(?P<upper>[^"]+)(?P<end>".*)$', re.MULTILINE)
    match = pattern.search(current)

    if not match:
        raise ApplicationError(f"Unable to find the 'requires' entry in: {ANSIBLE_PYPROJECT_TOML_FILE.relative_to(CHECKOUT_DIR)}")

    current_version = Version(match.group('upper'))

    if requested_version == current_version:
        return

    display.show(f"Updating setuptools upper bound from {current_version} to {requested_version} ...")

    updated = pattern.sub(fr'\g<begin>\g<lower>\g<middle>{requested_version}\g<end>', current)

    if current == updated:
        raise RuntimeError("Failed to set the setuptools upper bound.")

    ANSIBLE_PYPROJECT_TOML_FILE.write_text(updated)


def create_reproducible_sdist(original_path: pathlib.Path, output_path: pathlib.Path, mtime: int) -> None:
    """Read the specified sdist and write out a new copy with uniform file metadata at the specified location."""
    with tarfile.open(original_path) as original_archive:
        with tempfile.TemporaryDirectory() as temp_dir:
            tar_file = pathlib.Path(temp_dir) / "sdist.tar"

            with tarfile.open(tar_file, mode="w") as tar_archive:
                for original_info in original_archive.getmembers():  # type: tarfile.TarInfo
                    tar_archive.addfile(create_reproducible_tar_info(original_info, mtime), original_archive.extractfile(original_info))

            with tar_file.open("rb") as tar_archive:
                with gzip.GzipFile(output_path, "wb", mtime=mtime) as output_archive:
                    shutil.copyfileobj(tar_archive, output_archive)


def create_reproducible_tar_info(original: tarfile.TarInfo, mtime: int) -> tarfile.TarInfo:
    """Return a copy of the given TarInfo with uniform file metadata."""
    sanitized = tarfile.TarInfo()
    sanitized.name = original.name
    sanitized.size = original.size
    sanitized.mtime = mtime
    sanitized.mode = (original.mode & ~(stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)) | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWUSR
    sanitized.type = original.type
    sanitized.linkname = original.linkname
    sanitized.uid = 0
    sanitized.gid = 0
    sanitized.uname = "root"
    sanitized.gname = "root"

    if original.mode & (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH):
        sanitized.mode |= stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH

    return sanitized


def test_built_artifact(path: pathlib.Path) -> None:
    """Test the specified built artifact by installing it in a venv and running some basic commands."""
    with tempfile.TemporaryDirectory() as temp_dir_name:
        temp_dir = pathlib.Path(temp_dir_name)

        venv_dir = temp_dir / "venv"
        venv_bin_dir = venv_dir / "bin"

        venv.create(venv_dir, with_pip=True)

        env = os.environ.copy()
        env.pop("PYTHONPATH", None)  # avoid interference from ansible being injected into the environment
        env.update(
            PATH=os.pathsep.join((str(venv_bin_dir), env["PATH"])),
        )

        run("pip", "install", path, env=env | PIP_ENV, cwd=CHECKOUT_DIR)

        run("ansible", "--version", env=env, cwd=CHECKOUT_DIR)
        run("ansible-test", "--version", env=env, cwd=CHECKOUT_DIR)


def get_sdist_path(version: Version, dist_dir: pathlib.Path = DIST_DIR) -> pathlib.Path:
    """Return the path to the sdist file."""
    return dist_dir / f"ansible_core-{version}.tar.gz"


def get_wheel_path(version: Version, dist_dir: pathlib.Path = DIST_DIR) -> pathlib.Path:
    """Return the path to the wheel file."""
    return dist_dir / f"ansible_core-{version}-py3-none-any.whl"


def calculate_digest(path: pathlib.Path) -> str:
    """Return the digest for the specified file."""
    with open(path, "rb") as f:
        digest = hashlib.file_digest(f, DIGEST_ALGORITHM)
    return digest.hexdigest()


@functools.cache
def get_release_artifact_details(repository: str, version: Version, validate: bool) -> list[ReleaseArtifact]:
    """Return information about the release artifacts hosted on PyPI."""
    data = get_pypi_project(repository, 'ansible-core', version)
    artifacts = [describe_release_artifact(version, item, validate) for item in data["urls"]]

    expected_artifact_types = {"bdist_wheel", "sdist"}
    found_artifact_types = set(artifact.package_type for artifact in artifacts)

    if found_artifact_types != expected_artifact_types:
        raise RuntimeError(f"Expected {expected_artifact_types} artifact types, but found {found_artifact_types} instead.")

    return artifacts


def describe_release_artifact(version: Version, item: dict[str, t.Any], validate: bool) -> ReleaseArtifact:
    """Return release artifact details extracted from the given PyPI data."""
    package_type = item["packagetype"]

    # The artifact URL is documented as stable, so is safe to put in release notes.
    # See: https://github.com/pypi/warehouse/blame/c95be4a1055f4b36a8852715eb80318c81fc00ca/docs/api-reference/integration-guide.rst#L86-L90
    url = item["url"]

    pypi_size = item["size"]
    pypi_digest = item["digests"][DIGEST_ALGORITHM]

    if package_type == "bdist_wheel":
        local_artifact_file = get_wheel_path(version)
        package_label = "Built Distribution"
    elif package_type == "sdist":
        local_artifact_file = get_sdist_path(version)
        package_label = "Source Distribution"
    else:
        raise NotImplementedError(f"Package type '{package_type}' is not supported.")

    if validate:
        try:
            local_size = local_artifact_file.stat().st_size
            local_digest = calculate_digest(local_artifact_file)
        except FileNotFoundError:
            raise ApplicationError(f"Missing local artifact: {local_artifact_file.relative_to(CHECKOUT_DIR)}") from None

        if local_size != pypi_size:
            raise ApplicationError(f"The {version} local {package_type} size {local_size} does not match the PyPI size {pypi_size}.")

        if local_digest != pypi_digest:
            raise ApplicationError(f"The {version} local {package_type} digest '{local_digest}' does not match the PyPI digest '{pypi_digest}'.")

    return ReleaseArtifact(
        package_type=package_type,
        package_label=package_label,
        url=url,
        size=pypi_size,
        digest=pypi_digest,
        digest_algorithm=DIGEST_ALGORITHM.upper(),
    )


def get_next_release_date(start: datetime.date, step: int, after: datetime.date) -> datetime.date:
    """Return the next release date."""
    if start > after:
        raise ValueError(f"{start=} is greater than {after=}")

    current_delta = after - start
    release_delta = datetime.timedelta(days=(math.floor(current_delta.days / step) + 1) * step)

    release = start + release_delta

    return release


def create_template_environment() -> jinja2.Environment:
    """Create and return a jinja2 environment."""
    env = jinja2.Environment()
    env.filters.update(
        basename=os.path.basename,
    )

    return env


def create_github_release_notes(upstream: Remote, repository: str, version: Version, validate: bool) -> str:
    """Create and return GitHub release notes."""
    env = create_template_environment()
    template = env.from_string(GITHUB_RELEASE_NOTES_TEMPLATE)

    variables = dict(
        version=version,
        releases=get_release_artifact_details(repository, version, validate),
        changelog=f"https://github.com/{upstream.user}/{upstream.repo}/blob/v{version}/changelogs/CHANGELOG-v{version.major}.{version.minor}.rst",
    )

    release_notes = template.render(**variables).strip()

    return release_notes


# endregion
# region Templates


GITHUB_RELEASE_NOTES_TEMPLATE = """
# Changelog

See the [full changelog]({{ changelog }}) for the changes included in this release.

# Release Artifacts

{%- for release in releases %}
* {{ release.package_label }}: [{{ release.url|basename }}]({{ release.url }}) - &zwnj;{{ release.size }} bytes
  * {{ release.digest }} ({{ release.digest_algorithm }})
{%- endfor %}
"""


# endregion
# region Commands


command = CommandFramework(
    repository=dict(metavar="REPO", choices=tuple(PYPI_ENDPOINTS), default="pypi", help="PyPI repository to use: %(choices)s [%(default)s]"),
    version=dict(exclusive="version", help="version to set"),
    pre=dict(exclusive="version", help="increment version to the specified pre-release (aN, bN, rcN)"),
    final=dict(exclusive="version", action="store_true", help="increment version to the next final release"),
    commit=dict(help="commit to tag"),
    validate=dict(name="--no-validate", action="store_false", help="disable validation of PyPI artifacts against local ones"),
    prompt=dict(name="--no-prompt", action="store_false", help="disable interactive prompt before publishing with twine"),
    setuptools=dict(name='--no-setuptools', action="store_false", help="disable updating setuptools upper bound"),
    allow_tag=dict(action="store_true", help="allow an existing release tag (for testing)"),
    allow_stale=dict(action="store_true", help="allow a stale checkout (for testing)"),
    allow_dirty=dict(action="store_true", help="allow untracked files and files with changes (for testing)"),
)


@command
def instructions() -> None:
    """Show instructions for the release process."""
    message = """
Releases must be performed using an up-to-date checkout of a fork of the Ansible repository.

1. Make sure your checkout is up-to-date.
2. Run the `prepare` command [1], then:
   a. Submit the PR opened in the browser.
   b. Wait for CI to pass.
   c. Merge the PR.
3. Update your checkout to include the commit from the PR which was just merged.
4. Run the `complete` command [2], then:
   a. Submit the GitHub release opened in the browser.
   b. Submit the PR opened in the browser.
   c. Wait for CI to pass.
   d. Merge the PR.

[1] Use the `--final`, `--pre` or `--version` option for control over the version.
[2] During the `publish` step, `twine` may prompt for credentials.
"""

    display.show(message.strip())


@command
def show_version(final: bool = False, pre: str | None = None) -> None:
    """Show the current and next ansible-core version."""
    current_version = get_ansible_version(mode=VersionMode.ALLOW_DEV_POST)
    display.show(f"Current version: {current_version}")

    try:
        next_version = get_next_version(current_version, final=final, pre=pre)
    except ApplicationError as ex:
        display.show(f"   Next version: Unknown - {ex}")
    else:
        display.show(f"   Next version: {next_version}")

        check_ansible_version(current_version, next_version)


@command
def check_state(allow_stale: bool = False) -> None:
    """Verify the git repository is in a usable state for creating a pull request."""
    get_git_state(get_ansible_version(), allow_stale)


# noinspection PyUnusedLocal
@command
def prepare(final: bool = False, pre: str | None = None, version: str | None = None, setuptools: bool | None = None) -> None:
    """Prepare a release."""
    command.run(
        update_version,
        update_setuptools,
        check_state,
        generate_summary,
        generate_changelog,
        create_release_pr,
    )


@command
def update_version(final: bool = False, pre: str | None = None, version: str | None = None) -> None:
    """Update the version embedded in the source code."""
    current_version = get_ansible_version(mode=VersionMode.REQUIRE_DEV_POST)

    if version:
        requested_version = get_ansible_version(version)
    else:
        requested_version = get_next_version(current_version, final=final, pre=pre)

    set_ansible_version(current_version, requested_version)


@command
def update_setuptools(setuptools: bool) -> None:
    """Update the setuptools upper bound in pyproject.toml."""
    if not setuptools:
        return

    requested_version = get_latest_setuptools_version()
    set_setuptools_upper_bound(requested_version)


@command
def generate_summary() -> None:
    """Generate a summary changelog fragment for this release."""
    version = get_ansible_version()
    release_date = datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d")
    summary_path = CHANGELOGS_FRAGMENTS_DIR / f"{version}_summary.yaml"
    major_minor = f"{version.major}.{version.minor}"

    content = f"""
release_summary: |
   | Release Date: {release_date}
   | `Porting Guide <https://docs.ansible.com/ansible-core/{major_minor}/porting_guides/porting_guide_core_{major_minor}.html>`__
"""

    summary_path.write_text(content.lstrip())


@command
def generate_changelog() -> None:
    """Generate the changelog and validate the results."""
    changelog_requirements = (
        ANSIBLE_CHANGELOG_REQUIREMENTS_FILE.read_text()
        + ANSIBLE_REQUIREMENTS_FILE.read_text()  # TODO: consider pinning the ansible requirements and dependencies
    )

    env = ensure_venv(changelog_requirements)
    env.update(
        PATH=os.pathsep.join((str(ANSIBLE_BIN_DIR), env["PATH"])),
        PYTHONPATH=ANSIBLE_LIB_DIR,
    )

    # TODO: consider switching back to the original changelog generator instead of using antsibull-changelog

    run("antsibull-changelog", "release", "-vv", "--use-ansible-doc", env=env, cwd=CHECKOUT_DIR)
    run("antsibull-changelog", "generate", "-vv", "--use-ansible-doc", env=env, cwd=CHECKOUT_DIR)

    run("ansible-test", "sanity", CHANGELOGS_DIR, ANSIBLE_RELEASE_FILE, env=env, cwd=CHECKOUT_DIR)


@command
def create_release_pr(allow_stale: bool = False) -> None:
    """Create a branch and open a browser tab for creating a release pull request."""
    version = get_ansible_version()

    pr = prepare_pull_request(
        version=version,
        branch=f"release-{version}-{secrets.token_hex(4)}",
        title=f"New release v{version}",
        add=(
            CHANGELOGS_DIR,
            ANSIBLE_RELEASE_FILE,
            ANSIBLE_PYPROJECT_TOML_FILE,
        ),
        allow_stale=allow_stale,
    )

    create_pull_request(pr)


# noinspection PyUnusedLocal
@command
def complete(repository: str, allow_dirty: bool = False) -> None:
    """Complete a release after the prepared changes have been merged."""
    command.run(
        check_state,
        build,
        test,
        publish,
        tag_release,
        post_version,
        create_post_pr,
    )


@command
def build(allow_dirty: bool = False) -> None:
    """Build the sdist and wheel."""
    version = get_ansible_version(mode=VersionMode.ALLOW_DEV_POST)

    # TODO: consider pinning the build requirement and its dependencies
    build_requirements = """
build
"""
    env = ensure_venv(build_requirements)

    dirty = git("status", "--porcelain", "--untracked-files=all", capture_output=True).stdout.strip().splitlines()

    if dirty:
        with suppress_when(allow_dirty):
            raise ApplicationError(f"There are {len(dirty)} files which are untracked and/or have changes, which will be omitted from the build.")

    sdist_file = get_sdist_path(version)
    wheel_file = get_wheel_path(version)

    with tempfile.TemporaryDirectory(dir=DIST_DIR, prefix=f"build-{version}-", suffix=".tmp") as temp_dir_name:
        temp_dir = pathlib.Path(temp_dir_name)
        dist_dir = temp_dir / "dist"

        commit_time = int(git("show", "-s", "--format=%ct", capture_output=True).stdout)

        env.update(
            SOURCE_DATE_EPOCH=commit_time,
        )

        git("worktree", "add", "-d", temp_dir)

        try:
            run("python", "-m", "build", env=env, cwd=temp_dir)

            create_reproducible_sdist(get_sdist_path(version, dist_dir), sdist_file, commit_time)
            get_wheel_path(version, dist_dir).rename(wheel_file)
        finally:
            git("worktree", "remove", temp_dir)


@command
def test() -> None:
    """Test the sdist and wheel."""
    command.run(
        test_sdist,
        test_wheel,
    )


@command
def test_sdist() -> None:
    """Test the sdist."""
    version = get_ansible_version(mode=VersionMode.ALLOW_DEV_POST)
    sdist_file = get_sdist_path(version)

    with tempfile.TemporaryDirectory() as temp_dir_name:
        temp_dir = pathlib.Path(temp_dir_name)

        with contextlib.ExitStack() as stack:
            try:
                sdist = stack.enter_context(tarfile.open(sdist_file))
            except FileNotFoundError:
                raise ApplicationError(f"Missing sdist: {sdist_file.relative_to(CHECKOUT_DIR)}") from None

            # deprecated: description='extractall fallback without filter' python_version='3.11'
            if hasattr(tarfile, 'data_filter'):
                sdist.extractall(temp_dir, filter='data')  # type: ignore[call-arg]
            else:
                sdist.extractall(temp_dir)

        pyc_glob = "*.pyc*"
        pyc_files = sorted(path.relative_to(temp_dir) for path in temp_dir.rglob(pyc_glob))

        if pyc_files:
            raise ApplicationError(f"Found {len(pyc_files)} '{pyc_glob}' file(s): {', '.join(map(str, pyc_files))}")

    test_built_artifact(sdist_file)


@command
def test_wheel() -> None:
    """Test the wheel."""
    version = get_ansible_version(mode=VersionMode.ALLOW_DEV_POST)
    wheel_file = get_wheel_path(version)

    with tempfile.TemporaryDirectory() as temp_dir_name:
        temp_dir = pathlib.Path(temp_dir_name)

        with contextlib.ExitStack() as stack:
            try:
                wheel = stack.enter_context(zipfile.ZipFile(wheel_file))
            except FileNotFoundError:
                raise ApplicationError(f"Missing wheel for version {version}: {wheel_file}") from None

            wheel.extractall(temp_dir)

    test_built_artifact(wheel_file)


@command
def publish(repository: str, prompt: bool = True) -> None:
    """Publish to PyPI."""
    version = get_ansible_version()
    sdist_file = get_sdist_path(version)
    wheel_file = get_wheel_path(version)

    # TODO: consider pinning the twine requirement and its dependencies
    publish_requirements = """
twine
"""
    env = ensure_venv(publish_requirements)

    if prompt:
        try:
            while input(f"Do you want to publish {version} to the '{repository}' repository?\nEnter the repository name to confirm: ") != repository:
                pass
        except KeyboardInterrupt:
            display.show("")
            raise ApplicationError("Publishing was aborted by the user.") from None

    run("twine", "upload", "-r", repository, sdist_file, wheel_file, env=env, cwd=CHECKOUT_DIR)


@command
def tag_release(repository: str, commit: str | None = None, validate: bool = True, allow_tag: bool = False) -> None:
    """Create a GitHub release using the current or specified commit."""
    upstream = get_remotes().upstream

    if commit:
        git("fetch", upstream.name)  # fetch upstream to make sure the commit can be found

    commit = get_commit(commit)
    version = get_ansible_version(commit=commit)
    tag = f"v{version}"

    if upstream_tag := git("ls-remote", "--tags", upstream.name, tag, capture_output=True).stdout.strip():
        with suppress_when(allow_tag):
            raise ApplicationError(f"Version {version} has already been tagged: {upstream_tag}")

    upstream_branch = get_upstream_branch(version)
    upstream_refs = git("branch", "-r", "--format=%(refname)", "--contains", commit, capture_output=True).stdout.strip().splitlines()
    upstream_ref = f"refs/remotes/{upstream.name}/{upstream_branch}"

    if upstream_ref not in upstream_refs:
        raise ApplicationError(f"Commit {upstream_ref} not found. Found {len(upstream_refs)} upstream ref(s): {', '.join(upstream_refs)}")

    body = create_github_release_notes(upstream, repository, version, validate)

    release = GitHubRelease(
        user=upstream.user,
        repo=upstream.repo,
        target=commit,
        tag=tag,
        title=tag,
        body=body,
        pre_release=version.pre is not None,
    )

    create_github_release(release)


@command
def post_version() -> None:
    """Set the post release version."""
    current_version = get_ansible_version()
    requested_version = get_ansible_version(f"{current_version}.post0", mode=VersionMode.REQUIRE_POST)

    set_ansible_version(current_version, requested_version)


@command
def create_post_pr(allow_stale: bool = False) -> None:
    """Create a branch and open a browser tab for creating a post release pull request."""
    version = get_ansible_version(mode=VersionMode.REQUIRE_POST)

    pr = prepare_pull_request(
        version=version,
        branch=f"release-{version}-{secrets.token_hex(4)}",
        title=f"Update Ansible release version to v{version}.",
        add=(ANSIBLE_RELEASE_FILE,),
        allow_stale=allow_stale,
    )

    create_pull_request(pr)


# endregion


if __name__ == "__main__":
    command.main()



#!/usr/bin/env python

from __future__ import annotations

import argparse
import copy
import json
import os
import sys

inventory = dict(
    _meta=dict(hostvars=dict(localhost=dict(a_localhost_hostvar="avalue"), host1=dict(a_host1_hostvar="avalue"))),
    all=dict(children=["ungrouped", "group1", "empty_group"]),
    ungrouped=dict(hosts=["localhost"]),
    group1=dict(hosts=["host1"], vars=dict(a_group1_groupvar="avalue", group1_untrusted_var=dict(__ansible_unsafe="untrusted value"))),
    list_as_group=[],  # assigned to `hosts` in a new dict
    # this test case will be rewritten as a group with a host named `rewrite_as_host`, both vars will be untrusted
    rewrite_as_host=dict(avar="value", untrusted_var=dict(__ansible_unsafe="untrusted value")),
)

inventory_with_profile = copy.deepcopy(inventory)
inventory_with_profile['_meta']['profile'] = 'inventory_legacy'

normal_outputs = dict(
    # invalid
    invalid_utf8=b'i am invalid UTF-8 \xff\n',
    invalid_json="i am not JSON",
    invalid_type='"i am not a dict"',
    invalid_meta_type=dict(_meta='not a dict'),
    invalid_profile_type=dict(_meta=dict(profile=1)),
    invalid_profile_name=dict(_meta=dict(profile='invalid_profile')),
    invalid_inventory_profile_name=dict(_meta=dict(profile='inventory_invalid_profile')),
    invalid_json_for_profile=dict(dict(__ansible_vault=1)),
    invalid_meta_hostvars_type=dict(_meta=dict(hostvars=[])),
    invalid_meta_hostvars_type_for_host=dict(_meta=dict(hostvars=dict(myhost=[])), mygroup=dict(hosts=['myhost'])),
    invalid_group_vars_type=dict(_meta=dict(hostvars={}), mygroup=dict(vars=[])),
    invalid_group_type=dict(_meta=dict(hostvars={}), mygroup=None),
    no_meta_hostvars_host_nonzero_rc=dict(mygroup=dict(hosts=["myhost"])),  # trigger legacy "slow" path which invokes the script with --host for each host
    no_meta_hostvars_host_invalid_json=dict(mygroup=dict(hosts=["myhost"])),  # trigger legacy "slow" path which invokes the script with --host for each host
    # valid
    no_hosts=dict(_meta=dict(hostvars={})),
    no_profile=inventory,
    with_profile=inventory_with_profile,
    no_meta_hostvars=dict(mygroup=dict(hosts=["myhost"])),  # trigger legacy "slow" path which invokes the script with --host for each host
    no_meta_hostvars_empty_host_result=dict(mygroup=dict(hosts=["myhost"])),  # trigger legacy "slow" path which invokes the script with --host for each host
)

host_outputs = dict(
    # invalid/fail
    no_meta_hostvars_host_nonzero_rc=lambda: sys.exit(1),
    no_meta_hostvars_host_invalid_json=lambda: sys.exit(print("this is not JSON")),
    # valid
    no_meta_hostvars=dict(
        myhost=dict(avar="avalue", untrusted=dict(__ansible_unsafe="untrusted value")),
    ),
    no_meta_hostvars_empty_host_result=lambda: sys.exit(0),
)


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument('--host')
    parser.add_argument('--list', action='store_true')

    args = parser.parse_args()
    host: str | None = args.host

    if os.environ.get("INVENTORY_EMIT_STDERR"):
        print('this is stderr', file=sys.stderr, end='')

    if host:
        outputs = host_outputs
    else:
        outputs = normal_outputs

    if not (output := outputs.get(os.environ.get('INVENTORY_TEST_MODE'))):
        print(f"The `INVENTORY_TEST_MODE` envvar should be one of: \n{os.linesep.join(outputs)}", file=sys.stderr)
        sys.exit(1)

    if callable(output):
        output()

    if host:
        output = output.get(host)

    if isinstance(output, bytes):
        sys.stdout.buffer.write(output)
    else:
        if not isinstance(output, str):
            output = json.dumps(output)

        print(output)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# Most of these names are only available via PluginLoader so pylint doesn't
# know they exist
# pylint: disable=no-name-in-module
from __future__ import annotations

from ansible.module_utils.basic import AnsibleModule
import datetime

module = AnsibleModule(argument_spec=dict(
    datetime=dict(type=str, required=True),
    date=dict(type=str, required=True),
))
result = {
    'datetime': datetime.datetime.strptime(module.params.get('datetime'), '%Y-%m-%dT%H:%M:%S'),
    'date': datetime.datetime.strptime(module.params.get('date'), '%Y-%m-%d').date(),
}
module.exit_json(**result)



# Copyright (c) 2024 Ansible Project
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

"""Internal error handling logic for targets. Not for use on the controller."""

from __future__ import annotations as _annotations

import traceback as _sys_traceback
import typing as _t

from . import _messages

MSG_REASON_DIRECT_CAUSE: _t.Final[str] = '<<< caused by >>>'
MSG_REASON_HANDLING_CAUSE: _t.Final[str] = '<<< while handling >>>'

TRACEBACK_REASON_EXCEPTION_DIRECT_WARNING: _t.Final[str] = 'The above exception was the direct cause of the following warning:'


class EventFactory:
    """Factory for creating `Event` instances from `BaseException` instances on targets."""

    _MAX_DEPTH = 10
    """Maximum exception chain depth. Exceptions beyond this depth will be omitted."""

    @classmethod
    def from_exception(cls, exception: BaseException, include_traceback: bool) -> _messages.Event:
        return cls(include_traceback)._convert_exception(exception)

    def __init__(self, include_traceback: bool) -> None:
        self._include_traceback = include_traceback
        self._depth = 0

    def _convert_exception(self, exception: BaseException) -> _messages.Event:
        if self._depth > self._MAX_DEPTH:
            return _messages.Event(
                msg="Maximum depth exceeded, omitting further events.",
            )

        self._depth += 1

        try:
            return _messages.Event(
                msg=self._get_msg(exception),
                formatted_traceback=self._get_formatted_traceback(exception),
                formatted_source_context=self._get_formatted_source_context(exception),
                help_text=self._get_help_text(exception),
                chain=self._get_chain(exception),
                events=self._get_events(exception),
            )
        finally:
            self._depth -= 1

    def _get_msg(self, exception: BaseException) -> str | None:
        return str(exception).strip()

    def _get_formatted_traceback(self, exception: BaseException) -> str | None:
        if self._include_traceback:
            return ''.join(_sys_traceback.format_exception(type(exception), exception, exception.__traceback__, chain=False))

        return None

    def _get_formatted_source_context(self, exception: BaseException) -> str | None:
        return None

    def _get_help_text(self, exception: BaseException) -> str | None:
        return None

    def _get_chain(self, exception: BaseException) -> _messages.EventChain | None:
        if cause := self._get_cause(exception):
            return _messages.EventChain(
                msg_reason=MSG_REASON_DIRECT_CAUSE,
                traceback_reason='The above exception was the direct cause of the following exception:',
                event=self._convert_exception(cause),
                follow=self._follow_cause(exception),
            )

        if context := self._get_context(exception):
            return _messages.EventChain(
                msg_reason=MSG_REASON_HANDLING_CAUSE,
                traceback_reason='During handling of the above exception, another exception occurred:',
                event=self._convert_exception(context),
                follow=False,
            )

        return None

    def _follow_cause(self, exception: BaseException) -> bool:
        return True

    def _get_cause(self, exception: BaseException) -> BaseException | None:
        return exception.__cause__

    def _get_context(self, exception: BaseException) -> BaseException | None:
        if exception.__suppress_context__:
            return None

        return exception.__context__

    def _get_events(self, exception: BaseException) -> tuple[_messages.Event, ...] | None:
        # deprecated: description='move BaseExceptionGroup support here from ControllerEventFactory' python_version='3.10'
        return None



from __future__ import annotations

import typing as t

import pytest

from ansible.errors import AnsibleError
from ansible.module_utils.common._utils import get_all_subclasses
from ansible.module_utils._internal import _messages
from ansible._internal._templating._jinja_common import Marker, TruncationMarker, CapturedExceptionMarker, VaultExceptionMarker
from ansible._internal._templating._engine import TemplateEngine, TemplateOptions
from ansible._internal._templating._utils import TemplateContext


@pytest.fixture
def template_context() -> t.Iterator[TemplateContext]:
    """A fixture that provides a TemplateContext for the duration of a test."""
    with TemplateContext(template_value=None, templar=TemplateEngine(), options=TemplateOptions.DEFAULT, stop_on_template=False) as ctx:
        yield ctx


def get_concrete_marker_types() -> list[type[Marker]]:
    """Return a sorted list of Marker and its derived types."""
    return sorted(get_all_subclasses(Marker, include_abstract=False, consider_self=True), key=lambda au: au.__name__)


@pytest.fixture(params=get_concrete_marker_types())
def marker(request, template_context: TemplateContext) -> t.Iterator[Marker]:
    """
    A multiplying parameterized fixture that will yield an instance of each Marker-derived type.
    Depends on the template_context fixture, since these types can only be created under templating.
    """
    request_type = request.param

    if issubclass(request_type, TruncationMarker):
        yield request_type()
    elif issubclass(request_type, VaultExceptionMarker):
        yield VaultExceptionMarker(ciphertext='a ciphertext', event=_messages.Event(msg='a msg'))
    elif issubclass(request_type, CapturedExceptionMarker):
        try:
            try:
                raise Exception('bang')
            except Exception as ex:
                defer = ex

            raise AnsibleError('big bang') from defer  # pylint: disable=used-before-assignment  # false positive
        except Exception as ex2:
            yield request_type(ex2)
    else:
        yield request_type(hint="a hint", obj="obj", name="name")



from __future__ import annotations

import typing as t

import pytest

from pytest_mock import MockerFixture

from ansible._internal._errors._handler import ErrorHandler, ErrorAction
from ansible.parsing.vault import VaultSecretsContext, VaultSecret
from ansible._internal._templating._engine import _TemplateConfig
from units.mock.vault_helper import VaultTestHelper


# DTFIX-FUTURE: it'd be nice not to need to sync all controller-only fixtures here, but sunder values are excluded from `import *`,
#  and if we don't sunder the fixtures, they'll be unused args. Could also be handled with a managed import, module getattr, or others.
__all__ = ['_empty_vault_secrets_context', '_ignore_untrusted_template', '_vault_secrets_context', '_zap_vault_secrets_context']


@pytest.fixture
def _zap_vault_secrets_context() -> t.Generator[None]:
    """A fixture that disables any existing VaultSecretsContext both before and after the test."""
    # DTFIX-FUTURE: come up with a decent way to share this kind of fixture between controller contexts; we can't put it in top-level conftest since
    #  modules can't use it. Import from a shared controller fixture location, area-specific fixture support files, or ?
    VaultSecretsContext._current = None

    try:
        yield
    finally:
        VaultSecretsContext._current = None


@pytest.fixture
def _empty_vault_secrets_context(_zap_vault_secrets_context) -> t.Generator[None]:
    """A fixture that provides a `VaultSecretsContext` with no secrets."""
    VaultSecretsContext.initialize(VaultSecretsContext([]))

    yield


@pytest.fixture
def _vault_secrets_context(_zap_vault_secrets_context) -> t.Generator[VaultTestHelper]:
    """A fixture that initializes `VaultSecretsContext` with a single secret under the default ID and returns a `VaultTestHelper`."""
    secret = VaultSecret(b'secretbytesblah')

    VaultSecretsContext.initialize(VaultSecretsContext(secrets=[('default', secret)]))

    yield VaultTestHelper()


@pytest.fixture
def _ignore_untrusted_template(mocker: MockerFixture) -> t.Generator[None]:
    mocker.patch.object(_TemplateConfig, 'untrusted_template_handler', ErrorHandler(ErrorAction.IGNORE))

    yield



"""Unit tests to verify the functionality of the ansible-forked pytest plugin."""
from __future__ import annotations


import os
import pytest
import signal
import sys
import warnings


warnings.warn("This verifies that warnings generated during test collection are reported.")


@pytest.mark.xfail
def test_kill_xfail():
    os.kill(os.getpid(), signal.SIGKILL)  # causes pytest to report stdout and stderr


def test_kill():
    os.kill(os.getpid(), signal.SIGKILL)  # causes pytest to report stdout and stderr


@pytest.mark.xfail
def test_exception_xfail():
    sys.stdout.write("This stdout message should be hidden due to xfail.")
    sys.stderr.write("This stderr message should be hidden due to xfail.")
    raise Exception("This error is expected, but should be hidden due to xfail.")


def test_exception():
    sys.stdout.write("This stdout message should be reported since we're throwing an exception.")
    sys.stderr.write("This stderr message should be reported since we're throwing an exception.")
    raise Exception("This error is expected and should be visible.")


def test_warning():
    warnings.warn("This verifies that warnings generated at test time are reported.")


def test_passed():
    pass



# Copyright (c) 2022 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
    name: constructed_with_hostvars
    options:
      plugin:
        description: the load name of the plugin
      plugin_expression:
        description: an expression that must be trusted whose default resolves to 2
        default: 1 + 1
    extends_documentation_fragment:
      - constructed
      - fragment_with_expression
"""

from ansible.errors import AnsibleParserError
from ansible.plugins.inventory import BaseInventoryPlugin, Constructable
from ansible._internal import _testing


class InventoryModule(BaseInventoryPlugin, Constructable):

    NAME = 'constructed_with_hostvars'

    def verify_file(self, path):
        return super(InventoryModule, self).verify_file(path) and path.endswith(('constructed.yml', 'constructed.yaml'))

    def parse(self, inventory, loader, path, cache=True):
        super(InventoryModule, self).parse(inventory, loader, path, cache)
        config = self._read_config_data(path)

        with _testing.hard_fail_context("ensure config defaults are trusted and runnable as expressions") as ctx:
            ctx.check(self._compose(self.get_option('plugin_expression'), variables={}) == 2)
            ctx.check(self._compose(self.get_option('fragment_expression'), variables={}) == 4)

        strict = self.get_option('strict')
        try:
            for host in inventory.hosts:
                hostvars = {}

                # constructed groups based on conditionals
                self._add_host_to_composed_groups(self.get_option('groups'), hostvars, host, strict=strict, fetch_hostvars=True)

                # constructed groups based variable values
                self._add_host_to_keyed_groups(self.get_option('keyed_groups'), hostvars, host, strict=strict, fetch_hostvars=True)

        except Exception as ex:
            raise AnsibleParserError(f"Failed to parse {path}.") from ex



# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

from units.modules.utils import AnsibleExitJson, AnsibleFailJson, fail_json, exit_json
import pytest

from ansible.modules import iptables

IPTABLES_CMD = "/sbin/iptables"
IP6TABLES_CMD = "/sbin/ip6tables"
IPTABLES_VERSION = "1.8.2"
CONST_INPUT_FILTER = [IPTABLES_CMD, "-t", "filter", "-L", "INPUT",]


@pytest.fixture
def _mock_basic_commands(mocker):
    """Mock basic commands like get_bin_path and get_iptables_version."""
    mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.get_bin_path",
        return_value=IPTABLES_CMD,
    )
    mocker.patch("ansible.modules.iptables.get_iptables_version", return_value=IPTABLES_VERSION)


@pytest.mark.usefixtures('_mock_basic_commands')
def test_flush_table_without_chain(mocker, set_module_args):
    """Test flush without chain, flush the table."""
    set_module_args(
        {
            "flush": True,
        }
    )
    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command", return_value=(0, "", "")
    )
    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args[0][0]
    assert first_cmd_args_list[0], IPTABLES_CMD
    assert first_cmd_args_list[1], "-t"
    assert first_cmd_args_list[2], "filter"
    assert first_cmd_args_list[3], "-F"


@pytest.mark.usefixtures('_mock_basic_commands')
def test_flush_table_check_true(mocker, set_module_args):
    """Test flush without parameters and check == true."""
    set_module_args(
        {
            "flush": True,
            "_ansible_check_mode": True,
        }
    )
    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command", return_value=(0, "", "")
    )
    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 0


@pytest.mark.usefixtures('_mock_basic_commands')
def test_policy_table(mocker, set_module_args):
    """Test change policy of a chain."""
    set_module_args(
        {
            "policy": "ACCEPT",
            "chain": "INPUT",
        }
    )
    commands_results = [(0, "Chain INPUT (policy DROP)\n", ""), (0, "", "")]
    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )
    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 2
    first_cmd_args_list = run_command.call_args_list[0]
    second_cmd_args_list = run_command.call_args_list[1]
    assert first_cmd_args_list[0][0] == CONST_INPUT_FILTER
    assert second_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-P",
        "INPUT",
        "ACCEPT",
    ]


@pytest.mark.usefixtures('_mock_basic_commands')
@pytest.mark.parametrize(
    ("test_input", "commands_results"),
    [
        pytest.param(
            {
                "policy": "ACCEPT",
                "chain": "INPUT",
                "_ansible_check_mode": True,
            },
            [
                (0, "Chain INPUT (policy DROP)\n", ""),
            ],
            id="policy-table-no-change",
        ),
        pytest.param(
            {
                "policy": "ACCEPT",
                "chain": "INPUT",
            },
            [
                (0, "Chain INPUT (policy ACCEPT)\n", ""),
                (0, "", ""),
            ],
            id="policy-table-change-false",
        )
    ]
)
def test_policy_table_flush(mocker, test_input, commands_results, set_module_args):
    """Test flush without parameters and change == false."""
    set_module_args(test_input)
    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )
    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == CONST_INPUT_FILTER


# TODO ADD test policy without chain fail
# TODO ADD test policy with chain don't exists
# TODO ADD test policy with wrong choice fail

@pytest.mark.usefixtures('_mock_basic_commands')
def test_insert_rule_change_false(mocker, set_module_args):
    """Test flush without parameters."""
    set_module_args(
        {
            "chain": "OUTPUT",
            "source": "1.2.3.4/32",
            "destination": "7.8.9.10/42",
            "jump": "ACCEPT",
            "action": "insert",
            "_ansible_check_mode": True,
        }
    )
    commands_results = [
        (1, "", ""),  # check_rule_present
        (0, "", ""),  # check_chain_present
    ]
    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-C",
        "OUTPUT",
        "-s",
        "1.2.3.4/32",
        "-d",
        "7.8.9.10/42",
        "-j",
        "ACCEPT",
    ]


@pytest.mark.usefixtures('_mock_basic_commands')
def test_insert_rule(mocker, set_module_args):
    """Test flush without parameters."""
    set_module_args(
        {
            "chain": "OUTPUT",
            "source": "1.2.3.4/32",
            "destination": "7.8.9.10/42",
            "jump": "ACCEPT",
            "action": "insert",
        }
    )
    commands_results = [
        (1, "", ""),  # check_rule_present
        (0, "", ""),
    ]
    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 2
    first_cmd_args_list = run_command.call_args_list[0]
    second_cmd_args_list = run_command.call_args_list[1]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-C",
        "OUTPUT",
        "-s",
        "1.2.3.4/32",
        "-d",
        "7.8.9.10/42",
        "-j",
        "ACCEPT",
    ]
    assert second_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-I",
        "OUTPUT",
        "-s",
        "1.2.3.4/32",
        "-d",
        "7.8.9.10/42",
        "-j",
        "ACCEPT",
    ]


@pytest.mark.usefixtures('_mock_basic_commands')
def test_append_rule_check_mode(mocker, set_module_args):
    """Test append a redirection rule in check mode."""
    set_module_args(
        {
            "chain": "PREROUTING",
            "source": "1.2.3.4/32",
            "destination": "7.8.9.10/42",
            "jump": "REDIRECT",
            "table": "nat",
            "to_destination": "5.5.5.5/32",
            "protocol": "udp",
            "destination_port": "22",
            "to_ports": "8600",
            "_ansible_check_mode": True,
        }
    )

    commands_results = [
        (1, "", ""),  # check_rule_present
        (0, "", ""),  # check_chain_present
    ]
    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "nat",
        "-C",
        "PREROUTING",
        "-p",
        "udp",
        "-s",
        "1.2.3.4/32",
        "-d",
        "7.8.9.10/42",
        "-j",
        "REDIRECT",
        "--to-destination",
        "5.5.5.5/32",
        "--destination-port",
        "22",
        "--to-ports",
        "8600",
    ]


@pytest.mark.usefixtures('_mock_basic_commands')
def test_append_rule(mocker, set_module_args):
    """Test append a redirection rule."""
    set_module_args(
        {
            "chain": "PREROUTING",
            "source": "1.2.3.4/32",
            "destination": "7.8.9.10/42",
            "jump": "REDIRECT",
            "table": "nat",
            "to_destination": "5.5.5.5/32",
            "protocol": "udp",
            "destination_port": "22",
            "to_ports": "8600",
        }
    )

    commands_results = [
        (1, "", ""),  # check_rule_present
        (0, "", ""),  # check_chain_present
        (0, "", ""),
    ]
    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 2
    first_cmd_args_list = run_command.call_args_list[0]
    second_cmd_args_list = run_command.call_args_list[1]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "nat",
        "-C",
        "PREROUTING",
        "-p",
        "udp",
        "-s",
        "1.2.3.4/32",
        "-d",
        "7.8.9.10/42",
        "-j",
        "REDIRECT",
        "--to-destination",
        "5.5.5.5/32",
        "--destination-port",
        "22",
        "--to-ports",
        "8600",
    ]
    assert second_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "nat",
        "-A",
        "PREROUTING",
        "-p",
        "udp",
        "-s",
        "1.2.3.4/32",
        "-d",
        "7.8.9.10/42",
        "-j",
        "REDIRECT",
        "--to-destination",
        "5.5.5.5/32",
        "--destination-port",
        "22",
        "--to-ports",
        "8600",
    ]


@pytest.mark.usefixtures('_mock_basic_commands')
def test_remove_rule(mocker, set_module_args):
    """Test flush without parameters."""
    set_module_args(
        {
            "chain": "PREROUTING",
            "source": "1.2.3.4/32",
            "destination": "7.8.9.10/42",
            "jump": "SNAT",
            "table": "nat",
            "to_source": "5.5.5.5/32",
            "protocol": "udp",
            "source_port": "22",
            "to_ports": "8600",
            "state": "absent",
            "in_interface": "eth0",
            "out_interface": "eth1",
            "comment": "this is a comment",
        }
    )

    commands_results = [
        (0, "", ""),
        (0, "", ""),
    ]
    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 2
    first_cmd_args_list = run_command.call_args_list[0]
    second_cmd_args_list = run_command.call_args_list[1]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "nat",
        "-C",
        "PREROUTING",
        "-p",
        "udp",
        "-s",
        "1.2.3.4/32",
        "-d",
        "7.8.9.10/42",
        "-j",
        "SNAT",
        "--to-source",
        "5.5.5.5/32",
        "-i",
        "eth0",
        "-o",
        "eth1",
        "--source-port",
        "22",
        "--to-ports",
        "8600",
        "-m",
        "comment",
        "--comment",
        "this is a comment",
    ]
    assert second_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "nat",
        "-D",
        "PREROUTING",
        "-p",
        "udp",
        "-s",
        "1.2.3.4/32",
        "-d",
        "7.8.9.10/42",
        "-j",
        "SNAT",
        "--to-source",
        "5.5.5.5/32",
        "-i",
        "eth0",
        "-o",
        "eth1",
        "--source-port",
        "22",
        "--to-ports",
        "8600",
        "-m",
        "comment",
        "--comment",
        "this is a comment",
    ]


@pytest.mark.usefixtures('_mock_basic_commands')
def test_remove_rule_check_mode(mocker, set_module_args):
    """Test flush without parameters check mode."""
    set_module_args(
        {
            "chain": "PREROUTING",
            "source": "1.2.3.4/32",
            "destination": "7.8.9.10/42",
            "jump": "SNAT",
            "table": "nat",
            "to_source": "5.5.5.5/32",
            "protocol": "udp",
            "source_port": "22",
            "to_ports": "8600",
            "state": "absent",
            "in_interface": "eth0",
            "out_interface": "eth1",
            "comment": "this is a comment",
            "_ansible_check_mode": True,
        }
    )

    commands_results = [
        (0, "", ""),
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "nat",
        "-C",
        "PREROUTING",
        "-p",
        "udp",
        "-s",
        "1.2.3.4/32",
        "-d",
        "7.8.9.10/42",
        "-j",
        "SNAT",
        "--to-source",
        "5.5.5.5/32",
        "-i",
        "eth0",
        "-o",
        "eth1",
        "--source-port",
        "22",
        "--to-ports",
        "8600",
        "-m",
        "comment",
        "--comment",
        "this is a comment",
    ]


@pytest.mark.usefixtures('_mock_basic_commands')
@pytest.mark.parametrize(
    ("test_input", "expected"),
    [
        pytest.param(
            {
                "chain": "INPUT",
                "protocol": "tcp",
                "reject_with": "tcp-reset",
                "ip_version": "ipv4",
            },
            [
                IPTABLES_CMD,
                "-t",
                "filter",
                "-C",
                "INPUT",
                "-p",
                "tcp",
                "-j",
                "REJECT",
                "--reject-with",
                "tcp-reset",
            ],
            id="insert-reject-with",
        ),
        pytest.param(
            {
                "chain": "INPUT",
                "protocol": "tcp",
                "jump": "REJECT",
                "reject_with": "tcp-reset",
                "ip_version": "ipv4",
            },
            [
                IPTABLES_CMD,
                "-t",
                "filter",
                "-C",
                "INPUT",
                "-p",
                "tcp",
                "-j",
                "REJECT",
                "--reject-with",
                "tcp-reset",
            ],
            id="update-reject-with",
        ),
    ]
)
def test_insert_with_reject(mocker, test_input, expected, set_module_args):
    """Using reject_with with a previously defined jump: REJECT results in two Jump statements #18988."""
    set_module_args(test_input)
    commands_results = [
        (0, "", ""),
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == expected


@pytest.mark.usefixtures('_mock_basic_commands')
def test_jump_tee_gateway_negative(mocker, set_module_args):
    """Missing gateway when JUMP is set to TEE."""
    set_module_args(
        {
            "table": "mangle",
            "chain": "PREROUTING",
            "in_interface": "eth0",
            "protocol": "udp",
            "match": "state",
            "jump": "TEE",
            "ctstate": ["NEW"],
            "destination_port": "9521",
            "destination": "127.0.0.1",
        }
    )
    mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.fail_json",
        side_effect=fail_json,
    )
    jump_err_msg = "jump is TEE but all of the following are missing: gateway"
    with pytest.raises(AnsibleFailJson, match=jump_err_msg) as exc:
        iptables.main()
    assert exc.value.args[0]["failed"]


@pytest.mark.usefixtures('_mock_basic_commands')
def test_jump_tee_gateway(mocker, set_module_args):
    """Using gateway when JUMP is set to TEE."""
    set_module_args(
        {
            "table": "mangle",
            "chain": "PREROUTING",
            "in_interface": "eth0",
            "protocol": "udp",
            "match": "state",
            "jump": "TEE",
            "ctstate": ["NEW"],
            "destination_port": "9521",
            "gateway": "192.168.10.1",
            "destination": "127.0.0.1",
        }
    )
    commands_results = [
        (0, "", ""),
    ]
    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "mangle",
        "-C",
        "PREROUTING",
        "-p",
        "udp",
        "-d",
        "127.0.0.1",
        "-m",
        "state",
        "-j",
        "TEE",
        "--gateway",
        "192.168.10.1",
        "-i",
        "eth0",
        "--destination-port",
        "9521",
        "--state",
        "NEW",
    ]


@pytest.mark.usefixtures('_mock_basic_commands')
@pytest.mark.parametrize(
    ("test_input"),
    [
        pytest.param(
            'flags=ALL flags_set="ACK,RST,SYN,FIN"',
            id="tcp-flags-str"
        ),
        pytest.param(
            {
                "flags": "ALL", "flags_set": "ACK,RST,SYN,FIN"
            },
            id="tcp-flags-dict"
        ),
        pytest.param(
            {
                "flags": ["ALL"], "flags_set": ["ACK", "RST", "SYN", "FIN"]
            },
            id="tcp-flags-list"
        ),
    ],
)
def test_tcp_flags(mocker, test_input, set_module_args):
    """Test various ways of inputting tcp_flags."""
    rule_data = {
        "chain": "OUTPUT",
        "protocol": "tcp",
        "jump": "DROP",
        "tcp_flags": test_input,
    }

    set_module_args(rule_data)

    commands_results = [
        (0, "", ""),
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-C",
        "OUTPUT",
        "-p",
        "tcp",
        "--tcp-flags",
        "ALL",
        "ACK,RST,SYN,FIN",
        "-j",
        "DROP",
    ]


@pytest.mark.usefixtures('_mock_basic_commands')
@pytest.mark.parametrize(
    "log_level",
    [
        "0",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "emerg",
        "alert",
        "crit",
        "error",
        "warning",
        "notice",
        "info",
        "debug",
    ],
)
def test_log_level(mocker, log_level, set_module_args):
    """Test various ways of log level flag."""

    set_module_args(
        {
            "chain": "INPUT",
            "jump": "LOG",
            "log_level": log_level,
            "source": "1.2.3.4/32",
            "log_prefix": "** DROP-this_ip **",
        }
    )
    commands_results = [
        (0, "", ""),
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-C",
        "INPUT",
        "-s",
        "1.2.3.4/32",
        "-j",
        "LOG",
        "--log-prefix",
        "** DROP-this_ip **",
        "--log-level",
        log_level,
    ]


@pytest.mark.usefixtures('_mock_basic_commands')
@pytest.mark.parametrize(
    ("test_input", "expected"),
    [
        pytest.param(
            {
                "chain": "INPUT",
                "match": ["iprange"],
                "src_range": "192.168.1.100-192.168.1.199",
                "jump": "ACCEPT",
            },
            [
                IPTABLES_CMD,
                "-t",
                "filter",
                "-C",
                "INPUT",
                "-m",
                "iprange",
                "-j",
                "ACCEPT",
                "--src-range",
                "192.168.1.100-192.168.1.199",
            ],
            id="src-range",
        ),
        pytest.param(
            {
                "chain": "INPUT",
                "src_range": "192.168.1.100-192.168.1.199",
                "dst_range": "10.0.0.50-10.0.0.100",
                "jump": "ACCEPT",
            },
            [
                IPTABLES_CMD,
                "-t",
                "filter",
                "-C",
                "INPUT",
                "-j",
                "ACCEPT",
                "-m",
                "iprange",
                "--src-range",
                "192.168.1.100-192.168.1.199",
                "--dst-range",
                "10.0.0.50-10.0.0.100",
            ],
            id="src-range-dst-range",
        ),
        pytest.param(
            {
                "chain": "INPUT",
                "dst_range": "10.0.0.50-10.0.0.100",
                "jump": "ACCEPT"
            },
            [
                IPTABLES_CMD,
                "-t",
                "filter",
                "-C",
                "INPUT",
                "-j",
                "ACCEPT",
                "-m",
                "iprange",
                "--dst-range",
                "10.0.0.50-10.0.0.100",
            ],
            id="dst-range"
        ),
    ],
)
def test_iprange(mocker, test_input, expected, set_module_args):
    """Test iprange module with its flags src_range and dst_range."""
    set_module_args(test_input)

    commands_results = [
        (0, "", ""),
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == expected


@pytest.mark.usefixtures('_mock_basic_commands')
def test_insert_rule_with_wait(mocker, set_module_args):
    """Test flush without parameters."""
    set_module_args(
        {
            "chain": "OUTPUT",
            "source": "1.2.3.4/32",
            "destination": "7.8.9.10/42",
            "jump": "ACCEPT",
            "action": "insert",
            "wait": "10",
        }
    )

    commands_results = [
        (0, "", ""),
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-C",
        "OUTPUT",
        "-w",
        "10",
        "-s",
        "1.2.3.4/32",
        "-d",
        "7.8.9.10/42",
        "-j",
        "ACCEPT",
    ]


@pytest.mark.usefixtures('_mock_basic_commands')
def test_comment_position_at_end(mocker, set_module_args):
    """Test comment position to make sure it is at the end of command."""
    set_module_args(
        {
            "chain": "INPUT",
            "jump": "ACCEPT",
            "action": "insert",
            "ctstate": ["NEW"],
            "comment": "this is a comment",
            "_ansible_check_mode": True,
        }
    )

    commands_results = [
        (0, "", ""),
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-C",
        "INPUT",
        "-j",
        "ACCEPT",
        "-m",
        "conntrack",
        "--ctstate",
        "NEW",
        "-m",
        "comment",
        "--comment",
        "this is a comment",
    ]
    assert run_command.call_args[0][0][14] == "this is a comment"


@pytest.mark.usefixtures('_mock_basic_commands')
def test_destination_ports(mocker, set_module_args):
    """Test multiport module usage with multiple ports."""
    set_module_args(
        {
            "chain": "INPUT",
            "protocol": "tcp",
            "in_interface": "eth0",
            "source": "192.168.0.1/32",
            "destination_ports": ["80", "443", "8081:8085"],
            "jump": "ACCEPT",
            "comment": "this is a comment",
        }
    )
    commands_results = [
        (0, "", ""),
    ]
    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-C",
        "INPUT",
        "-p",
        "tcp",
        "-s",
        "192.168.0.1/32",
        "-j",
        "ACCEPT",
        "-m",
        "multiport",
        "--dports",
        "80,443,8081:8085",
        "-i",
        "eth0",
        "-m",
        "comment",
        "--comment",
        "this is a comment",
    ]


@pytest.mark.usefixtures('_mock_basic_commands')
@pytest.mark.parametrize(
    ("test_input", "expected"),
    [
        pytest.param(
            {
                "chain": "INPUT",
                "protocol": "tcp",
                "match_set": "admin_hosts",
                "match_set_flags": "src",
                "destination_port": "22",
                "jump": "ACCEPT",
                "comment": "this is a comment",
            },
            [
                IPTABLES_CMD,
                "-t",
                "filter",
                "-C",
                "INPUT",
                "-p",
                "tcp",
                "-j",
                "ACCEPT",
                "--destination-port",
                "22",
                "-m",
                "set",
                "--match-set",
                "admin_hosts",
                "src",
                "-m",
                "comment",
                "--comment",
                "this is a comment",
            ],
            id="match-set-src",
        ),
        pytest.param(
            {
                "chain": "INPUT",
                "protocol": "udp",
                "match_set": "banned_hosts",
                "match_set_flags": "src,dst",
                "jump": "REJECT",
            },
            [
                IPTABLES_CMD,
                "-t",
                "filter",
                "-C",
                "INPUT",
                "-p",
                "udp",
                "-j",
                "REJECT",
                "-m",
                "set",
                "--match-set",
                "banned_hosts",
                "src,dst",
            ],
            id="match-set-src-dst",
        ),
        pytest.param(
            {
                "chain": "INPUT",
                "protocol": "udp",
                "match_set": "banned_hosts_dst",
                "match_set_flags": "dst,dst",
                "jump": "REJECT",
            },
            [
                IPTABLES_CMD,
                "-t",
                "filter",
                "-C",
                "INPUT",
                "-p",
                "udp",
                "-j",
                "REJECT",
                "-m",
                "set",
                "--match-set",
                "banned_hosts_dst",
                "dst,dst",
            ],
            id="match-set-dst-dst",
        ),
        pytest.param(
            {
                "chain": "INPUT",
                "protocol": "udp",
                "match_set": "banned_hosts",
                "match_set_flags": "src,src",
                "jump": "REJECT",
            },
            [
                IPTABLES_CMD,
                "-t",
                "filter",
                "-C",
                "INPUT",
                "-p",
                "udp",
                "-j",
                "REJECT",
                "-m",
                "set",
                "--match-set",
                "banned_hosts",
                "src,src",
            ],
            id="match-set-src-src",
        ),
    ],
)
def test_match_set(mocker, test_input, expected, set_module_args):
    """Test match_set together with match_set_flags."""
    set_module_args(test_input)
    commands_results = [
        (0, "", ""),
    ]
    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == expected


@pytest.mark.usefixtures('_mock_basic_commands')
def test_chain_creation(mocker, set_module_args):
    """Test chain creation when absent."""
    set_module_args(
        {
            "chain": "FOOBAR",
            "state": "present",
            "chain_management": True,
            "wait": 10,
        }
    )

    commands_results = [
        (1, "", ""),  # check_chain_present
        (0, "", ""),  # create_chain
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.exit_json",
        side_effect=exit_json,
    )
    with pytest.raises(AnsibleExitJson):
        iptables.main()

    assert run_command.call_count == 2
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-L",
        "FOOBAR",
        "-w",
        "10",
    ]

    second_cmd_args_list = run_command.call_args_list[1]
    assert second_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-N",
        "FOOBAR",
        "-w",
        "10",
    ]

    commands_results = [
        (0, "", ""),  # check_rule_present
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(AnsibleExitJson) as exc:
        iptables.main()
    assert not exc.value.args[0]["changed"]


@pytest.mark.usefixtures('_mock_basic_commands')
def test_chain_creation_check_mode(mocker, set_module_args):
    """Test chain creation when absent in check mode."""
    set_module_args(
        {
            "chain": "FOOBAR",
            "state": "present",
            "chain_management": True,
            "wait": 10,
            "_ansible_check_mode": True,
        }
    )

    commands_results = [
        (1, "", ""),  # check_rule_present
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.exit_json",
        side_effect=exit_json,
    )
    with pytest.raises(AnsibleExitJson):
        iptables.main()

    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-L",
        "FOOBAR",
        "-w",
        "10",
    ]

    commands_results = [
        (0, "", ""),  # check_rule_present
    ]
    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )
    with pytest.raises(AnsibleExitJson) as exc:
        iptables.main()

    assert not exc.value.args[0]["changed"]


@pytest.mark.usefixtures('_mock_basic_commands')
def test_chain_deletion(mocker, set_module_args):
    """Test chain deletion when present."""
    set_module_args(
        {
            "chain": "FOOBAR",
            "state": "absent",
            "chain_management": True,
        }
    )

    commands_results = [
        (0, "", ""),  # check_chain_present
        (0, "", ""),  # delete_chain
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.exit_json",
        side_effect=exit_json,
    )
    with pytest.raises(AnsibleExitJson) as exc:
        iptables.main()

    assert exc.value.args[0]["changed"]
    assert run_command.call_count == 2
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-L",
        "FOOBAR",
    ]
    second_cmd_args_list = run_command.call_args_list[1]
    assert second_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-X",
        "FOOBAR",
    ]

    commands_results = [
        (1, "", ""),  # check_rule_present
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(AnsibleExitJson) as exc:
        iptables.main()

    assert not exc.value.args[0]["changed"]


@pytest.mark.usefixtures('_mock_basic_commands')
def test_chain_deletion_check_mode(mocker, set_module_args):
    """Test chain deletion when present in check mode."""
    set_module_args(
        {
            "chain": "FOOBAR",
            "state": "absent",
            "chain_management": True,
            "_ansible_check_mode": True,
        }
    )

    commands_results = [
        (0, "", ""),  # check_chain_present
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.exit_json",
        side_effect=exit_json,
    )
    with pytest.raises(AnsibleExitJson) as exc:
        iptables.main()

    assert exc.value.args[0]["changed"]
    assert run_command.call_count == 1
    first_cmd_args_list = run_command.call_args_list[0]
    assert first_cmd_args_list[0][0] == [
        IPTABLES_CMD,
        "-t",
        "filter",
        "-L",
        "FOOBAR",
    ]

    commands_results = [
        (1, "", ""),  # check_rule_present
    ]

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(AnsibleExitJson) as exc:
        iptables.main()

    assert not exc.value.args[0]["changed"]


@pytest.mark.parametrize("ip_version", ["both"])
def test_ip_version(mocker, ip_version, set_module_args):
    """Test 'both' option for ip_version."""

    set_module_args(
        {
            "chain": "INPUT",
            "protocol": "tcp",
            "destination_port": "80",
            "jump": "ACCEPT",
            "ip_version": ip_version,
        }
    )

    commands_results = [
        (0, "", ""),
        (0, "", ""),
    ]

    def get_bin_path_side_effect(arg, *args, **kwargs):
        if arg == "iptables":
            return IPTABLES_CMD
        elif arg == "ip6tables":
            return IP6TABLES_CMD
        return None

    mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.get_bin_path",
        side_effect=get_bin_path_side_effect,
    )

    mocker.patch("ansible.modules.iptables.get_iptables_version", return_value=IPTABLES_VERSION)

    run_command = mocker.patch(
        "ansible.module_utils.basic.AnsibleModule.run_command",
        side_effect=commands_results,
    )

    with pytest.raises(SystemExit):
        iptables.main()

    assert run_command.call_count == 2

    first_cmd_args = run_command.call_args_list[0][0][0]
    assert first_cmd_args == [
        IPTABLES_CMD,
        "-t", "filter",
        "-C", "INPUT",
        "-p", "tcp",
        "-j", "ACCEPT",
        "--destination-port", "80",
    ]

    second_cmd_args = run_command.call_args_list[1][0][0]
    assert second_cmd_args == [
        IP6TABLES_CMD,
        "-t", "filter",
        "-C", "INPUT",
        "-p", "tcp",
        "-j", "ACCEPT",
        "--destination-port", "80",
    ]



"""Classes for storing and processing test results."""

from __future__ import annotations

import collections.abc as c
import datetime
import typing as t

from .util import (
    display,
)

from .util_common import (
    get_docs_url,
    write_text_test_results,
    write_json_test_results,
    ResultType,
)

from .metadata import (
    Metadata,
)

from .config import (
    TestConfig,
)

from . import junit_xml


def calculate_best_confidence(choices: tuple[tuple[str, int], ...], metadata: Metadata) -> int:
    """Return the best confidence value available from the given choices and metadata."""
    best_confidence = 0

    for path, line in choices:
        confidence = calculate_confidence(path, line, metadata)
        best_confidence = max(confidence, best_confidence)

    return best_confidence


def calculate_confidence(path: str, line: int, metadata: Metadata) -> int:
    """Return the confidence level for a test result associated with the given file path and line number."""
    ranges = metadata.changes.get(path)

    # no changes were made to the file
    if not ranges:
        return 0

    # changes were made to the same file and line
    if any(r[0] <= line <= r[1] for r in ranges):
        return 100

    # changes were made to the same file and the line number is unknown
    if line == 0:
        return 75

    # changes were made to the same file and the line number is different
    return 50


class TestResult:
    """Base class for test results."""

    def __init__(self, command: str, test: str, python_version: t.Optional[str] = None) -> None:
        self.command = command
        self.test = test
        self.python_version = python_version
        self.name = self.test or self.command

        if self.python_version:
            self.name += '-python-%s' % self.python_version

    def write(self, args: TestConfig) -> None:
        """Write the test results to various locations."""
        self.write_console()
        self.write_bot(args)

        if args.lint:
            self.write_lint()

        if args.junit:
            self.write_junit(args)

    def write_console(self) -> None:
        """Write results to console."""

    def write_lint(self) -> None:
        """Write lint results to stdout."""

    def write_bot(self, args: TestConfig) -> None:
        """Write results to a file for ansibullbot to consume."""

    def write_junit(self, args: TestConfig) -> None:
        """Write results to a junit XML file."""

    def create_result_name(self, extension: str) -> str:
        """Return the name of the result file using the given extension."""
        name = 'ansible-test-%s' % self.command

        if self.test:
            name += '-%s' % self.test

        if self.python_version:
            name += '-python-%s' % self.python_version

        name += extension

        return name

    def save_junit(self, args: TestConfig, test_case: junit_xml.TestCase) -> None:
        """Save the given test case results to disk as JUnit XML."""
        suites = junit_xml.TestSuites(
            suites=[
                junit_xml.TestSuite(
                    name='ansible-test',
                    cases=[test_case],
                    timestamp=datetime.datetime.now(tz=datetime.timezone.utc),
                ),
            ],
        )

        report = suites.to_pretty_xml()

        if args.explain:
            return

        write_text_test_results(ResultType.JUNIT, self.create_result_name('.xml'), report)


class TestTimeout(TestResult):
    """Test timeout."""

    def __init__(self, timeout_duration: int | float) -> None:
        super().__init__(command='timeout', test='')

        self.timeout_duration = timeout_duration

    def write(self, args: TestConfig) -> None:
        """Write the test results to various locations."""
        message = 'Tests were aborted after exceeding the %d minute time limit.' % self.timeout_duration

        # Include a leading newline to improve readability on Shippable "Tests" tab.
        # Without this, the first line becomes indented.
        output = """
One or more of the following situations may be responsible:

- Code changes have resulted in tests that hang or run for an excessive amount of time.
- Tests have been added which exceed the time limit when combined with existing tests.
- Test infrastructure and/or external dependencies are operating slower than normal."""

        if args.coverage:
            output += '\n- Additional overhead from collecting code coverage has resulted in tests exceeding the time limit.'

        output += '\n\nConsult the console log for additional details on where the timeout occurred.'

        suites = junit_xml.TestSuites(
            suites=[
                junit_xml.TestSuite(
                    name='ansible-test',
                    timestamp=datetime.datetime.now(tz=datetime.timezone.utc),
                    cases=[
                        junit_xml.TestCase(
                            name='timeout',
                            classname='timeout',
                            errors=[
                                junit_xml.TestError(
                                    message=message,
                                ),
                            ],
                        ),
                    ],
                )
            ],
        )

        report = suites.to_pretty_xml()

        write_text_test_results(ResultType.JUNIT, self.create_result_name('.xml'), report)


class TestSuccess(TestResult):
    """Test success."""

    def write_junit(self, args: TestConfig) -> None:
        """Write results to a junit XML file."""
        test_case = junit_xml.TestCase(classname=self.command, name=self.name)

        self.save_junit(args, test_case)


class TestSkipped(TestResult):
    """Test skipped."""

    def __init__(self, command: str, test: str, python_version: t.Optional[str] = None) -> None:
        super().__init__(command, test, python_version)

        self.reason: t.Optional[str] = None

    def write_console(self) -> None:
        """Write results to console."""
        if self.reason:
            display.warning(self.reason)
        else:
            display.info('No tests applicable.', verbosity=1)

    def write_junit(self, args: TestConfig) -> None:
        """Write results to a junit XML file."""
        test_case = junit_xml.TestCase(
            classname=self.command,
            name=self.name,
            skipped=self.reason or 'No tests applicable.',
        )

        self.save_junit(args, test_case)


class TestFailure(TestResult):
    """Test failure."""

    def __init__(
        self,
        command: str,
        test: str,
        python_version: t.Optional[str] = None,
        messages: t.Optional[c.Sequence[TestMessage]] = None,
        summary: t.Optional[str] = None,
    ):
        super().__init__(command, test, python_version)

        if messages:
            messages = sorted(messages)
        else:
            messages = []

        self.messages = messages
        self.summary = summary

    def write(self, args: TestConfig) -> None:
        """Write the test results to various locations."""
        if args.metadata.changes:
            self.populate_confidence(args.metadata)

        super().write(args)

    def write_console(self) -> None:
        """Write results to console."""
        if self.summary:
            display.error(self.summary)
        else:
            if self.python_version:
                specifier = ' on python %s' % self.python_version
            else:
                specifier = ''

            display.error('Found %d %s issue(s)%s which need to be resolved:' % (len(self.messages), self.test or self.command, specifier))

            for message in self.messages:
                display.error(message.format(show_confidence=True))

            doc_url = self.find_docs()
            if doc_url:
                display.info('See documentation for help: %s' % doc_url)

    def write_lint(self) -> None:
        """Write lint results to stdout."""
        if self.summary:
            command = self.format_command()
            message = 'The test `%s` failed. See stderr output for details.' % command
            path = ''
            message = TestMessage(message, path)
            print(message)  # display goes to stderr, this should be on stdout
        else:
            for message in self.messages:
                print(message)  # display goes to stderr, this should be on stdout

    def write_junit(self, args: TestConfig) -> None:
        """Write results to a junit XML file."""
        title = self.format_title()
        output = self.format_block()

        test_case = junit_xml.TestCase(
            classname=self.command,
            name=self.name,
            failures=[
                junit_xml.TestFailure(
                    message=title,
                    output=output,
                ),
            ],
        )

        self.save_junit(args, test_case)

    def write_bot(self, args: TestConfig) -> None:
        """Write results to a file for ansibullbot to consume."""
        docs = self.find_docs()
        message = self.format_title(help_link=docs)
        output = self.format_block()

        if self.messages:
            verified = all((m.confidence or 0) >= 50 for m in self.messages)
        else:
            verified = False

        bot_data = dict(
            verified=verified,
            docs=docs,
            results=[
                dict(
                    message=message,
                    output=output,
                ),
            ],
        )

        if args.explain:
            return

        write_json_test_results(ResultType.BOT, self.create_result_name('.json'), bot_data)

    def populate_confidence(self, metadata: Metadata) -> None:
        """Populate test result confidence using the provided metadata."""
        for message in self.messages:
            if message.confidence is None:
                message.confidence = calculate_confidence(message.path, message.line, metadata)

    def format_command(self) -> str:
        """Return a string representing the CLI command associated with the test failure."""
        command = 'ansible-test %s' % self.command

        if self.test:
            command += ' --test %s' % self.test

        if self.python_version:
            command += ' --python %s' % self.python_version

        return command

    def find_docs(self) -> t.Optional[str]:
        """Return the docs URL for this test or None if there is no docs URL."""
        if self.command != 'sanity':
            return None  # only sanity tests have docs links

        filename = f'{self.test}.html' if self.test else ''
        url = get_docs_url(f'https://docs.ansible.com/ansible-core/devel/dev_guide/testing/{self.command}/{filename}')

        return url

    def format_title(self, help_link: t.Optional[str] = None) -> str:
        """Return a string containing a title/heading for this test failure, including an optional help link to explain the test."""
        command = self.format_command()

        if self.summary:
            reason = 'the error'
        else:
            reason = '1 error' if len(self.messages) == 1 else '%d errors' % len(self.messages)

        if help_link:
            help_link_markup = ' [[explain](%s)]' % help_link
        else:
            help_link_markup = ''

        title = 'The test `%s`%s failed with %s:' % (command, help_link_markup, reason)

        return title

    def format_block(self) -> str:
        """Format the test summary or messages as a block of text and return the result."""
        if self.summary:
            block = self.summary
        else:
            block = '\n'.join(m.format() for m in self.messages)

        message = block.strip()

        # Hack to remove ANSI color reset code from SubprocessError messages.
        message = message.replace(display.clear, '')

        return message


class TestMessage:
    """Single test message for one file."""

    def __init__(
        self,
        message: str,
        path: str,
        line: int = 0,
        column: int = 0,
        level: str = 'error',
        code: t.Optional[str] = None,
        confidence: t.Optional[int] = None,
    ):
        self.__path = path
        self.__line = line
        self.__column = column
        self.__level = level
        self.__code = code
        self.__message = message

        self.confidence = confidence

    @property
    def path(self) -> str:
        """Return the path."""
        return self.__path

    @property
    def line(self) -> int:
        """Return the line number, or 0 if none is available."""
        return self.__line

    @property
    def column(self) -> int:
        """Return the column number, or 0 if none is available."""
        return self.__column

    @property
    def level(self) -> str:
        """Return the level."""
        return self.__level

    @property
    def code(self) -> t.Optional[str]:
        """Return the code, if any."""
        return self.__code

    @property
    def message(self) -> str:
        """Return the message."""
        return self.__message

    @property
    def tuple(self) -> tuple[str, int, int, str, t.Optional[str], str]:
        """Return a tuple with all the immutable values of this test message."""
        return self.__path, self.__line, self.__column, self.__level, self.__code, self.__message

    def __lt__(self, other):
        return self.tuple < other.tuple

    def __le__(self, other):
        return self.tuple <= other.tuple

    def __eq__(self, other):
        return self.tuple == other.tuple

    def __ne__(self, other):
        return self.tuple != other.tuple

    def __gt__(self, other):
        return self.tuple > other.tuple

    def __ge__(self, other):
        return self.tuple >= other.tuple

    def __hash__(self):
        return hash(self.tuple)

    def __str__(self):
        return self.format()

    def format(self, show_confidence: bool = False) -> str:
        """Return a string representation of this message, optionally including the confidence level."""
        if self.__code:
            msg = '%s: %s' % (self.__code, self.__message)
        else:
            msg = self.__message

        if show_confidence and self.confidence is not None:
            msg += ' (%d%%)' % self.confidence

        return '%s:%s:%s: %s' % (self.__path, self.__line, self.__column, msg)



from __future__ import annotations

from ansible.plugins.action import ActionBase


class ActionModule(ActionBase):
    supports_raw_params = True

    def run(self, tmp=None, task_vars=None):
        action_args = self._task.args
        return dict(action_args=action_args)



"""
Utilities to support unit testing of Ansible Python modules.
Not supported for use cases other than testing.
"""

from __future__ import annotations as _annotations

import contextlib as _contextlib
import json as _json
import typing as _t

from unittest import mock as _mock

from ansible.module_utils.common import json as _common_json
from . import basic as _basic


@_contextlib.contextmanager
def patch_module_args(args: dict[str, _t.Any] | None = None) -> _t.Iterator[None]:
    """Expose the given module args to `AnsibleModule` instances created within this context."""
    if not isinstance(args, (dict, type(None))):
        raise TypeError("The `args` arg must be a dict or None.")

    args = dict(ANSIBLE_MODULE_ARGS=args or {})
    profile = 'legacy'  # this should be configurable in the future, once the profile feature is more fully baked

    encoder = _common_json.get_module_encoder(profile, _common_json.Direction.CONTROLLER_TO_MODULE)
    args = _json.dumps(args, cls=encoder).encode()

    with _mock.patch.object(_basic, '_ANSIBLE_ARGS', args), _mock.patch.object(_basic, '_ANSIBLE_PROFILE', profile):
        yield



from __future__ import annotations

import abc
import collections.abc as _c
import dataclasses
import itertools
import pathlib
import textwrap
import typing as t

from ansible._internal._datatag._tags import Origin
from ansible._internal._errors import _error_factory
from ansible.module_utils._internal import _ambient_context, _event_utils, _messages, _traceback


class ContributesToTaskResult(metaclass=abc.ABCMeta):
    """Exceptions may include this mixin to contribute task result dictionary data directly to the final result."""

    @property
    @abc.abstractmethod
    def result_contribution(self) -> _c.Mapping[str, object]:
        """Mapping of results to apply to the task result."""

    @property
    def omit_exception_key(self) -> bool:
        """Non-error exceptions (e.g., `AnsibleActionSkip`) must return `True` to ensure omission of the `exception` key."""
        return False

    @property
    def omit_failed_key(self) -> bool:
        """Exceptions representing non-failure scenarios (e.g., `skipped`, `unreachable`) must return `True` to ensure omisson of the `failed` key."""
        return False


class RedactAnnotatedSourceContext(_ambient_context.AmbientContextBase):
    """When active, this context will redact annotated source lines, showing only the origin."""


@dataclasses.dataclass(kw_only=True, frozen=True)
class SourceContext:
    origin: Origin
    annotated_source_lines: list[str]
    target_line: str | None

    def __str__(self) -> str:
        msg_lines = [f'Origin: {self.origin}']

        if self.annotated_source_lines:
            msg_lines.append('')
            msg_lines.extend(self.annotated_source_lines)

        return '\n'.join(msg_lines)

    @classmethod
    def from_value(cls, value: t.Any) -> SourceContext | None:
        """Attempt to retrieve source and render a contextual indicator from the value's origin (if any)."""
        if value is None:
            return None

        if isinstance(value, Origin):
            origin = value
            value = None
        else:
            origin = Origin.get_tag(value)

        if RedactAnnotatedSourceContext.current(optional=True):
            return cls.error('content redacted')

        if origin and origin.path:
            return cls.from_origin(origin)

        if value is None:
            truncated_value = None
            annotated_source_lines = []
        else:
            # DTFIX-FUTURE: cleanup/share width
            try:
                value = str(value)
            except Exception as ex:
                value = f'<< context unavailable: {ex} >>'

            truncated_value = textwrap.shorten(value, width=120)
            annotated_source_lines = [truncated_value]

        return SourceContext(
            origin=origin or Origin.UNKNOWN,
            annotated_source_lines=annotated_source_lines,
            target_line=truncated_value,
        )

    @staticmethod
    def error(message: str | None, origin: Origin | None = None) -> SourceContext:
        return SourceContext(
            origin=origin,
            annotated_source_lines=[f'(source not shown: {message})'] if message else [],
            target_line=None,
        )

    @classmethod
    def from_origin(cls, origin: Origin) -> SourceContext:
        """Attempt to retrieve source and render a contextual indicator of an error location."""
        from ansible.parsing.vault import is_encrypted  # avoid circular import

        # DTFIX-FUTURE: support referencing the column after the end of the target line, so we can indicate where a missing character (quote) needs to be added
        #               this is also useful for cases like end-of-stream reported by the YAML parser

        # DTFIX-FUTURE: Implement line wrapping and match annotated line width to the terminal display width.

        context_line_count: t.Final = 2
        max_annotated_line_width: t.Final = 120
        truncation_marker: t.Final = '...'

        target_line_num = origin.line_num

        if RedactAnnotatedSourceContext.current(optional=True):
            return cls.error('content redacted', origin)

        if not target_line_num or target_line_num < 1:
            return cls.error(None, origin)  # message omitted since lack of line number is obvious from pos

        start_line_idx = max(0, (target_line_num - 1) - context_line_count)  # if near start of file
        target_col_num = origin.col_num

        try:
            with pathlib.Path(origin.path).open() as src:
                first_line = src.readline()
                lines = list(itertools.islice(itertools.chain((first_line,), src), start_line_idx, target_line_num))
        except Exception as ex:
            return cls.error(type(ex).__name__, origin)

        if is_encrypted(first_line):
            return cls.error('content encrypted', origin)

        if len(lines) != target_line_num - start_line_idx:
            return cls.error('file truncated', origin)

        annotated_source_lines = []

        line_label_width = len(str(target_line_num))
        max_src_line_len = max_annotated_line_width - line_label_width - 1

        usable_line_len = max_src_line_len

        for line_num, line in enumerate(lines, start_line_idx + 1):
            line = line.rstrip('\n')  # universal newline default mode on `open` ensures we'll never see anything but \n
            line = line.replace('\t', ' ')  # mixed tab/space handling is intentionally disabled since we're both format and display config agnostic

            if len(line) > max_src_line_len:
                line = line[: max_src_line_len - len(truncation_marker)] + truncation_marker
                usable_line_len = max_src_line_len - len(truncation_marker)

            annotated_source_lines.append(f'{str(line_num).rjust(line_label_width)}{" " if line else ""}{line}')

        if target_col_num and usable_line_len >= target_col_num >= 1:
            column_marker = f'column {target_col_num}'

            target_col_idx = target_col_num - 1

            if target_col_idx + 2 + len(column_marker) > max_src_line_len:
                column_marker = f'{" " * (target_col_idx - len(column_marker) - 1)}{column_marker} ^'
            else:
                column_marker = f'{" " * target_col_idx}^ {column_marker}'

            column_marker = f'{" " * line_label_width} {column_marker}'

            annotated_source_lines.append(column_marker)
        elif target_col_num is None:
            underline_length = len(annotated_source_lines[-1]) - line_label_width - 1
            annotated_source_lines.append(f'{" " * line_label_width} {"^" * underline_length}')

        return SourceContext(
            origin=origin,
            annotated_source_lines=annotated_source_lines,
            target_line=lines[-1].rstrip('\n'),  # universal newline default mode on `open` ensures we'll never see anything but \n
        )


def format_exception_message(exception: BaseException) -> str:
    """Return the full chain of exception messages by concatenating the cause(s) until all are exhausted."""
    return _event_utils.format_event_brief_message(_error_factory.ControllerEventFactory.from_exception(exception, False))


def result_dict_from_exception(exception: BaseException, accept_result_contribution: bool = False) -> dict[str, object]:
    """Return a failed task result dict from the given exception."""
    event = _error_factory.ControllerEventFactory.from_exception(exception, _traceback.is_traceback_enabled(_traceback.TracebackEvent.ERROR))

    result: dict[str, object] = {}
    omit_failed_key = False
    omit_exception_key = False

    if accept_result_contribution:
        while exception:
            if isinstance(exception, ContributesToTaskResult):
                result = dict(exception.result_contribution)
                omit_failed_key = exception.omit_failed_key
                omit_exception_key = exception.omit_exception_key
                break

            exception = exception.__cause__

    if omit_failed_key:
        result.pop('failed', None)
    else:
        result.update(failed=True)

    if omit_exception_key:
        result.pop('exception', None)
    else:
        result.update(exception=_messages.ErrorSummary(event=event))

    if 'msg' not in result:
        # if nothing contributed `msg`, generate one from the exception messages
        result.update(msg=_event_utils.format_event_brief_message(event))

    return result


def result_dict_from_captured_errors(
    msg: str,
    *,
    errors: list[_messages.ErrorSummary] | None = None,
) -> dict[str, object]:
    """Return a failed task result dict from the given error message and captured errors."""
    _skip_stackwalk = True

    event = _messages.Event(
        msg=msg,
        formatted_traceback=_traceback.maybe_capture_traceback(msg, _traceback.TracebackEvent.ERROR),
        events=tuple(error.event for error in errors) if errors else None,
    )

    result = dict(
        failed=True,
        exception=_messages.ErrorSummary(
            event=event,
        ),
        msg=_event_utils.format_event_brief_message(event),
    )

    return result



# -*- coding: utf-8 -*-
# (c) 2018 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

from ansible.module_utils.common.sys_info import get_all_subclasses


#
# Tests for get_all_subclasses
#

class TestGetAllSubclasses:
    class Base:
        pass

    class BranchI(Base):
        pass

    class BranchII(Base):
        pass

    class BranchIA(BranchI):
        pass

    class BranchIB(BranchI):
        pass

    class BranchIIA(BranchII):
        pass

    class BranchIIB(BranchII):
        pass

    class MultipleInheritanceBase:
        pass

    class MultipleInheritanceBranchI(MultipleInheritanceBase):
        pass

    class MultipleInheritanceBranchII(MultipleInheritanceBase):
        pass

    class MultipleInheritanceChild(MultipleInheritanceBranchI, MultipleInheritanceBranchII):
        pass

    def test_bottom_level(self):
        assert get_all_subclasses(self.BranchIIB) == set()

    def test_one_inheritance(self):
        assert set(get_all_subclasses(self.BranchII)) == set([self.BranchIIA, self.BranchIIB])

    def test_toplevel(self):
        assert set(get_all_subclasses(self.Base)) == set([self.BranchI, self.BranchII,
                                                          self.BranchIA, self.BranchIB,
                                                          self.BranchIIA, self.BranchIIB])

    def test_multiple_inheritance(self) -> None:
        assert get_all_subclasses(self.MultipleInheritanceBase) == {self.MultipleInheritanceBranchI,
                                                                    self.MultipleInheritanceBranchII,
                                                                    self.MultipleInheritanceChild}



from __future__ import annotations

import os

import pytest
import pytest_mock

from ansible.constants import config
from ansible.errors import AnsibleUndefinedConfigEntry
from ansible._internal._errors._handler import ErrorHandler, ErrorAction, Skippable, _SkipException
from ansible.utils.display import Display


def test_skippable_ignore_skips_body() -> None:
    """Verify that `skip_on_ignore=True` skips the body within the context manager when `action=ErrorAction.IGNORE`."""
    body_ran = False
    assert not body_ran  # satisfy static analysis which assumes the context manager body will run

    with Skippable, ErrorHandler(ErrorAction.IGNORE).handle(Exception, skip_on_ignore=True):
        body_ran = True

    assert not body_ran


def test_skippable_without_skip_on_ignore() -> None:
    """
    Verify using `Skippable` without invoking a handler with `skip_on_ignore=True` will fail.
    This protects against accidental use of `Skippable` by itself, or forgetting to use `skip_on_ignore=True` -- both of which have no effect.
    """
    body_ran = False
    assert not body_ran  # satisfy static analysis which assumes the context manager body will run

    with pytest.raises(RuntimeError) as err:
        with Skippable:
            body_ran = True

    assert body_ran
    assert 'handler was never invoked' in str(err.value)


def test_skippable_non_skip_exception() -> None:
    """Verify that `Skippable` does not interfere with exceptions."""
    ex_to_raise = RuntimeError('let me through')

    with pytest.raises(RuntimeError) as err:
        with Skippable:
            raise ex_to_raise

    assert err.value is ex_to_raise


@pytest.mark.parametrize("error_action", (ErrorAction.IGNORE, ErrorAction.WARNING, ErrorAction.ERROR))
def test_skip_on_ignore_missing_skippable(error_action: ErrorAction) -> None:
    """Verify that a `_SkipException` is raised when `skip_on_ignore=True` and no `Skippable` context was used to suppress it."""
    body_ran = False
    assert not body_ran  # satisfy static analysis which assumes the context manager body will run

    with pytest.raises(_SkipException):
        with ErrorHandler(error_action).handle(Exception, skip_on_ignore=True):
            body_ran = True

    if error_action is ErrorAction.IGNORE:
        assert not body_ran
    else:
        assert body_ran


@pytest.mark.parametrize("exception_type", (RuntimeError, NotImplementedError))
def test_ignore_success(exception_type: type[Exception]) -> None:
    """Verify that `ErrorAction.IGNORE` suppresses the specified exception types."""
    body_ran = False
    assert not body_ran  # satisfy static analysis which assumes the context manager body will run

    with ErrorHandler(ErrorAction.IGNORE).handle(RuntimeError, NotImplementedError):
        body_ran = True
        raise exception_type('should be ignored')

    assert body_ran


def test_ignore_passes_other_exceptions() -> None:
    """Verify that `ErrorAction.IGNORE` does not suppress exception types not passed to `handle`."""
    with pytest.raises(NotImplementedError):
        with ErrorHandler(ErrorAction.IGNORE).handle(TypeError, ValueError):
            raise NotImplementedError()


@pytest.mark.parametrize("exception_type", (RuntimeError, NotImplementedError))
def test_warn_success(exception_type: type[Exception], mocker: pytest_mock.MockerFixture) -> None:
    """Verify that `ErrorAction.WARNING` eats the specified error type and calls `error_as_warning` with the exception instance raised."""
    eaw = mocker.patch.object(Display(), 'error_as_warning')
    with ErrorHandler(ErrorAction.WARNING).handle(RuntimeError, NotImplementedError):
        raise exception_type()

    assert isinstance(eaw.call_args.kwargs['exception'], exception_type)


def test_warn_passes_other_exceptions(mocker: pytest_mock.MockerFixture) -> None:
    """Verify that `ErrorAction.WARNING` does not suppress exception types not passed to `handle`, and that `error_as_warning` is not called for them."""
    eaw = mocker.patch.object(Display(), 'error_as_warning')

    with pytest.raises(NotImplementedError):
        with ErrorHandler(ErrorAction.WARNING).handle(TypeError, ValueError):
            raise NotImplementedError()

    assert not eaw.called


@pytest.mark.parametrize("exception_type", (AttributeError, NotImplementedError, ValueError))
def test_fail(exception_type: type[Exception]) -> None:
    """Verify that `ErrorAction.ERROR` passes through all exception types, regardless of what was passed to `handle`."""
    with pytest.raises(exception_type):
        with ErrorHandler(ErrorAction.ERROR).handle(AttributeError, NotImplementedError):
            raise exception_type()


def test_no_exceptions_to_handle():
    """Verify that passing no exceptions to `handle` fails."""
    with pytest.raises(ValueError):
        with ErrorHandler(ErrorAction.IGNORE).handle():
            pass


@pytest.mark.parametrize("value", ('ignore', 'warning', 'error'))
def test_from_config_env_success(value: str, mocker: pytest_mock.MockerFixture) -> None:
    """Verify that `from_config` correctly creates handlers with the requested error action config string."""
    mocker.patch.dict(os.environ, dict(_ANSIBLE_CALLBACK_DISPATCH_ERROR_BEHAVIOR=value))

    assert config.get_config_value("_CALLBACK_DISPATCH_ERROR_BEHAVIOR") == value
    eh = ErrorHandler.from_config("_CALLBACK_DISPATCH_ERROR_BEHAVIOR")
    assert eh.action == ErrorAction[value.upper()]


def test_from_config_fail() -> None:
    """Verify that `from_config` fails on an invalid config entry name."""
    with pytest.raises(AnsibleUndefinedConfigEntry):
        ErrorHandler.from_config("invalid")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# this is a windows documentation stub.  actual code lives in the .ps1
# file of the same name

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['stableinterface'],
                    'supported_by': 'core'}

DOCUMENTATION = r"""
---
module: win_stat
version_added: "1.7"
short_description: Get information about Windows files
description:
     - Returns information about a Windows file.
     - For non-Windows targets, use the M(stat) module instead.
options:
    path:
        description:
            - The full path of the file/object to get the facts of; both forward and
              back slashes are accepted.
        type: path
        required: yes
        aliases: [ dest, name ]
    get_checksum:
        description:
            - Whether to return a checksum of the file (default sha1)
        type: bool
        default: yes
        version_added: "2.1"
    checksum_algorithm:
        description:
            - Algorithm to determine checksum of file.
            - Will throw an error if the host is unable to use specified algorithm.
        type: str
        default: sha1
        choices: [ md5, sha1, sha256, sha384, sha512 ]
        version_added: "2.3"
    follow:
        description:
            - Whether to follow symlinks or junction points.
            - In the case of C(path) pointing to another link, then that will
              be followed until no more links are found.
        type: bool
        default: no
        version_added: "2.8"
seealso:
- module: stat
- module: win_acl
- module: win_file
- module: win_owner
author:
- Chris Church (@cchurch)
"""

EXAMPLES = r"""
- name: Obtain information about a file
  win_stat:
    path: C:\foo.ini
  register: file_info

- name: Obtain information about a folder
  win_stat:
    path: C:\bar
  register: folder_info

- name: Get MD5 checksum of a file
  win_stat:
    path: C:\foo.ini
    get_checksum: yes
    checksum_algorithm: md5
  register: md5_checksum

- debug:
    var: md5_checksum.stat.checksum

- name: Get SHA1 checksum of file
  win_stat:
    path: C:\foo.ini
    get_checksum: yes
  register: sha1_checksum

- debug:
    var: sha1_checksum.stat.checksum

- name: Get SHA256 checksum of file
  win_stat:
    path: C:\foo.ini
    get_checksum: yes
    checksum_algorithm: sha256
  register: sha256_checksum

- debug:
    var: sha256_checksum.stat.checksum
"""

RETURN = r"""
changed:
    description: Whether anything was changed
    returned: always
    type: bool
    sample: true
stat:
    description: dictionary containing all the stat data
    returned: success
    type: complex
    contains:
        attributes:
            description: Attributes of the file at path in raw form.
            returned: success, path exists
            type: str
            sample: "Archive, Hidden"
        checksum:
            description: The checksum of a file based on checksum_algorithm specified.
            returned: success, path exist, path is a file, get_checksum == True
              checksum_algorithm specified is supported
            type: str
            sample: 09cb79e8fc7453c84a07f644e441fd81623b7f98
        creationtime:
            description: The create time of the file represented in seconds since epoch.
            returned: success, path exists
            type: float
            sample: 1477984205.15
        exists:
            description: If the path exists or not.
            returned: success
            type: bool
            sample: true
        extension:
            description: The extension of the file at path.
            returned: success, path exists, path is a file
            type: str
            sample: ".ps1"
        filename:
            description: The name of the file (without path).
            returned: success, path exists, path is a file
            type: str
            sample: foo.ini
        hlnk_targets:
            description: List of other files pointing to the same file (hard links), excludes the current file.
            returned: success, path exists
            type: list
            sample:
            - C:\temp\file.txt
            - C:\Windows\update.log
        isarchive:
            description: If the path is ready for archiving or not.
            returned: success, path exists
            type: bool
            sample: true
        isdir:
            description: If the path is a directory or not.
            returned: success, path exists
            type: bool
            sample: true
        ishidden:
            description: If the path is hidden or not.
            returned: success, path exists
            type: bool
            sample: true
        isjunction:
            description: If the path is a junction point or not.
            returned: success, path exists
            type: bool
            sample: true
        islnk:
            description: If the path is a symbolic link or not.
            returned: success, path exists
            type: bool
            sample: true
        isreadonly:
            description: If the path is read only or not.
            returned: success, path exists
            type: bool
            sample: true
        isreg:
            description: If the path is a regular file.
            returned: success, path exists
            type: bool
            sample: true
        isshared:
            description: If the path is shared or not.
            returned: success, path exists
            type: bool
            sample: true
        lastaccesstime:
            description: The last access time of the file represented in seconds since epoch.
            returned: success, path exists
            type: float
            sample: 1477984205.15
        lastwritetime:
            description: The last modification time of the file represented in seconds since epoch.
            returned: success, path exists
            type: float
            sample: 1477984205.15
        lnk_source:
            description: Target of the symlink normalized for the remote filesystem.
            returned: success, path exists and the path is a symbolic link or junction point
            type: str
            sample: C:\temp\link
        lnk_target:
            description: Target of the symlink. Note that relative paths remain relative.
            returned: success, path exists and the path is a symbolic link or junction point
            type: str
            sample: ..\link
        nlink:
            description: Number of links to the file (hard links).
            returned: success, path exists
            type: int
            sample: 1
        owner:
            description: The owner of the file.
            returned: success, path exists
            type: str
            sample: BUILTIN\Administrators
        path:
            description: The full absolute path to the file.
            returned: success, path exists, file exists
            type: str
            sample: C:\foo.ini
        sharename:
            description: The name of share if folder is shared.
            returned: success, path exists, file is a directory and isshared == True
            type: str
            sample: file-share
        size:
            description: The size in bytes of a file or folder.
            returned: success, path exists, file is not a link
            type: int
            sample: 1024
"""



from __future__ import annotations

import typing as t

import pytest

from ansible.module_utils.common.yaml import yaml_dump, yaml_dump_all
from ansible.module_utils._internal._datatag import Tripwire
from ansible.module_utils._internal._datatag._tags import Deprecated
from units.mock.custom_types import CustomMapping, CustomSequence

_test_tag = Deprecated(msg="test")


@pytest.mark.parametrize("value, expected", (
    (CustomMapping(dict(a=1)), "a: 1"),
    (CustomSequence([1]), "- 1"),
    (_test_tag.tag(dict(a=1)), "a: 1"),
    (_test_tag.tag([1]), "- 1"),
    (_test_tag.tag(1), "1"),
    (_test_tag.tag("Ansible"), "Ansible"),
))
def test_dump(value: t.Any, expected: str) -> None:
    """Verify supported types can be dumped."""
    result = yaml_dump(value).strip()

    assert result == expected

    result = yaml_dump_all([value]).strip()

    assert result == expected


def test_dump_tripwire() -> None:
    """Verify dumping a tripwire trips it."""
    class Tripped(Exception):
        pass

    class CustomTripwire(Tripwire):
        def trip(self) -> t.NoReturn:
            raise Tripped()

    with pytest.raises(Tripped):
        yaml_dump(CustomTripwire())

    with pytest.raises(Tripped):
        yaml_dump_all([CustomTripwire()])



# Collect facts related to selinux
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import typing as t

from ansible.module_utils.facts.collector import BaseFactCollector

try:
    from ansible.module_utils.compat import selinux
    HAVE_SELINUX = True
except ImportError:
    HAVE_SELINUX = False

SELINUX_MODE_DICT = {
    1: 'enforcing',
    0: 'permissive',
    -1: 'disabled'
}


class SelinuxFactCollector(BaseFactCollector):
    name = 'selinux'
    _fact_ids = set()  # type: t.Set[str]

    def collect(self, module=None, collected_facts=None):
        facts_dict = {}
        selinux_facts = {}

        # If selinux library is missing, only set the status and selinux_python_present since
        # there is no way to tell if SELinux is enabled or disabled on the system
        # without the library.
        if not HAVE_SELINUX:
            selinux_facts['status'] = 'Missing selinux Python library'
            facts_dict['selinux'] = selinux_facts
            facts_dict['selinux_python_present'] = False
            return facts_dict

        # Set a boolean for testing whether the Python library is present
        facts_dict['selinux_python_present'] = True

        if not selinux.is_selinux_enabled():
            selinux_facts['status'] = 'disabled'
        else:
            selinux_facts['status'] = 'enabled'

            try:
                selinux_facts['policyvers'] = selinux.security_policyvers()
            except (AttributeError, OSError):
                selinux_facts['policyvers'] = 'unknown'

            try:
                (rc, configmode) = selinux.selinux_getenforcemode()
                if rc == 0:
                    selinux_facts['config_mode'] = SELINUX_MODE_DICT.get(configmode, 'unknown')
                else:
                    selinux_facts['config_mode'] = 'unknown'
            except (AttributeError, OSError):
                selinux_facts['config_mode'] = 'unknown'

            try:
                mode = selinux.security_getenforce()
                selinux_facts['mode'] = SELINUX_MODE_DICT.get(mode, 'unknown')
            except (AttributeError, OSError):
                selinux_facts['mode'] = 'unknown'

            try:
                (rc, policytype) = selinux.selinux_getpolicytype()
                if rc == 0:
                    selinux_facts['type'] = policytype
                else:
                    selinux_facts['type'] = 'unknown'
            except (AttributeError, OSError):
                selinux_facts['type'] = 'unknown'

        facts_dict['selinux'] = selinux_facts
        return facts_dict



# Copyright (c) 2018, Ansible Project
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

from __future__ import annotations

import os

from ansible.module_utils.common.file import is_executable
from ansible.module_utils.common.warnings import deprecate


def get_bin_path(arg, opt_dirs=None, required=None):
    """
    Find system executable in PATH. Raises ValueError if the executable is not found.

    :param arg: the executable to find
    :type arg: string
    :param opt_dirs: optional list of directories to search in addition to PATH
    :type opt_dirs: list of strings
    :param required: DEPRECATED. This parameter will be removed in 2.21
    :type required: boolean
    :returns: path to arg (should be abs path unless PATH or opt_dirs are relative paths)
    :raises: ValueError: if arg is not found

    In addition to PATH and opt_dirs, this function also looks through /sbin, /usr/sbin and /usr/local/sbin. A lot of
    modules, especially for gathering facts, depend on this behaviour.
    """
    if required is not None:
        deprecate(
            msg="The `required` parameter in `get_bin_path` API is deprecated.",
            version="2.21",
        )

    paths = []
    sbin_paths = ['/sbin', '/usr/sbin', '/usr/local/sbin']
    opt_dirs = [] if opt_dirs is None else opt_dirs

    # Construct possible paths with precedence
    # passed in paths
    for d in opt_dirs:
        if d is not None and os.path.exists(d):
            paths.append(d)
    # system configured paths
    paths += os.environ.get('PATH', '').split(os.pathsep)

    # existing /sbin dirs, if not there already
    for p in sbin_paths:
        if p not in paths and os.path.exists(p):
            paths.append(p)

    # Search for binary
    bin_path = None
    for d in paths:
        if not d:
            continue
        path = os.path.join(d, arg)
        if os.path.exists(path) and not os.path.isdir(path) and is_executable(path):
            # fist found wins
            bin_path = path
            break

    if bin_path is None:
        raise ValueError('Failed to find required executable "%s" in paths: %s' % (arg, os.pathsep.join(paths)))

    return bin_path



from __future__ import annotations

DOCUMENTATION = """
    name: dummy_cache
    short_description: dummy cache
    options:
      dummy:
        type: str
"""

import json
import os
import pathlib
import typing as t

from ansible.plugins.cache import BaseCacheModule


class CacheModule(BaseCacheModule):
    def __init__(self, *args, **kwargs):
        if not os.environ.get('DUMMY_CACHE_SKIP_SUPER'):
            super().__init__(*args, **kwargs)

        self._storage_dir = pathlib.Path(os.environ.get('OUTPUT_DIR')) / 'cache-storage'
        self._storage_dir.mkdir(parents=True, exist_ok=True)

    def _get_key_path(self, key: str) -> pathlib.Path:
        return self._storage_dir / key

    def get(self, key: t.Any) -> object:
        try:
            return json.loads(self._get_key_path(key).read_text())
        except FileNotFoundError:
            raise KeyError(key) from None

    def set(self, key: str, value: object) -> None:
        self._get_key_path(key).write_text(json.dumps(value))

    def keys(self) -> list[str]:
        path: pathlib.Path

        return [path.name for path in self._storage_dir.iterdir()]

    def contains(self, key: object) -> bool:
        try:
            self.get(key)
        except KeyError:
            return False

        return True

    def delete(self, key: str) -> None:
        self._get_key_path(key).unlink(missing_ok=True)

    def flush(self) -> None:
        for key in self.keys():
            self.delete(key)
